"""A generic class testing several views with different users.

The expected response can by defined by the response code, a HttpResponse, an exception or a callable.
In these case, a RequestTester will be built with this value as first argument.
Otherwise, you can also provide a RequestTester (especially to provide GET or POST arguments).
"""

import logging
from typing import Dict, List, Optional, Tuple, Type, TypeAlias, Union
from urllib.parse import urlencode

from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import site as default_site
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, AnonymousUser
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import PermissionDenied
from django.db import models
from django.http import Http404, HttpResponse, SimpleCookie
from django.middleware.csrf import get_token
from django.test import RequestFactory, TestCase
from django.urls import URLPattern
from django.urls.resolvers import RoutePattern
from django.utils.encoding import iri_to_uri

from df_site.testing.requests import RequestTester

logger = logging.getLogger(__name__)
ExpectedResponse: TypeAlias = Union[int, Exception, Type[Exception], HttpResponse, callable, RequestTester]
ExpectedResponses = Union[List[ExpectedResponse], ExpectedResponse]


class TestMultipleViews(TestCase):
    """Define some methods for testing a set of views with different users (identified by its username).

    `expected_responses` is a dict such that:
    - the key is the view name
    - the value is a dict such that:
        - the key is a USER_KEY (can be anything, like its privilege level)
        - the value is any of:
            - an integer (the expected HTTP status code — then this is the only checked thing)
            - an exception (only the type of the raised exception is checked)
            - a HttpResponse (only status_codes are checked)
            - a callable that will be given the TestCase and the HttpResponse as arguments
            - a RequestTester to deeply customize the test (including the request args/kwargs)
            - a list of any of the previous values

    You must override `get_user_keys` to provide a list of values to create different users.
        (`str(value)` is used to get the username)
    Your test_* methods must call
        check_url_patterns(url_patterns, common_kwargs, expected_responses, display_prefix=display_prefix)
    """

    views_test_counter = 0
    call_test_counter = 0

    @classmethod
    def setUpClass(cls):
        """Create some users and a request factory for testing."""
        super().setUpClass()
        # Every test needs access to the request factory.
        cls.request_factory = RequestFactory()
        cls.created_users: List[AbstractUser] = cls.create_users()

    def setUp(self):
        """Counts the number of run tests."""
        TestMultipleViews.call_test_counter += 1
        return super().setUp()

    @classmethod
    def get_users(cls) -> List[AbstractUser]:
        """Teturn a list of values to identify different users.

        The special value `None` is dedicated to AnonymousUser.
        """
        return cls.created_users

    @classmethod
    def create_users(cls):
        """Create some users for testing."""
        return [
            AnonymousUser(),
            cls.create_user("admin", is_staff=True, is_superuser=True),
            cls.create_user("staff", is_staff=True, is_superuser=False),
        ]

    @classmethod
    def create_user(cls, name, is_active=True, **kwargs):
        """Create a user with the provided name."""
        user = get_user_model().objects.filter(username=name).first()
        if user is not None:
            return user
        user_kwargs = {
            "email": f"{name}@test.com",
            "is_active": is_active,
        }
        user_kwargs.update(kwargs)
        return get_user_model().objects.create_user(username=name, **user_kwargs)

    def create_request(
        self,
        url,
        *args,
        method=None,
        user=None,
        form_data=None,
        get_data=None,
        post_data=None,
        **kwargs,
    ):
        """Return a request object for the provided URL."""
        if form_data:
            kwargs["data"] = urlencode(form_data)
            kwargs["content_type"] = "application/x-www-form-urlencoded"
        if method:
            method = method.lower()
        if method and method in ("get", "head") and get_data:
            kwargs["data"] = get_data
        elif get_data:
            kwargs["QUERY_STRING"] = urlencode(get_data, doseq=True)
        if post_data:
            kwargs["data"] = post_data
            method = method or "post"
            request = self.request_factory.generic(method.upper(), url, *args, **kwargs)
        else:
            method = method or "get"
            request = getattr(self.request_factory, method)(url, *args, **kwargs)
        request.user = user or AnonymousUser()
        request.cookies = SimpleCookie()
        setattr(request, "_dont_enforce_csrf_checks", True)
        request.cookies["csrftoken"] = get_token(request)

        def get_response():
            pass

        SessionMiddleware(get_response).process_request(request)
        MessageMiddleware(get_response).process_request(request)
        return request

    def check_url_patterns(
        self,
        url_patterns: List[Tuple[str, URLPattern]],
        common_kwargs: Dict[str, str],
        expected_responses: Dict[str, Dict[Optional[str], ExpectedResponses]],
        display_prefix: str = "",
    ):
        """Check all provided URL patterns.

        :param url_patterns: list of ("namespace:", URLPattern)  (leave "" if no namespace)
        :param common_kwargs: kwargs common to all URL patterns
            (use `expected_responses` to provided kwargs specific to a URL pattern)
        :param expected_responses:
        :param display_prefix: prefix added to displayed error messages
        :return:
        """
        failures = []  # list of (url_pattern, ps, exception)
        for namespace, url_pattern in url_patterns:
            if url_pattern.name:
                view_name = f"{namespace}{url_pattern.name}"
            else:
                view_name = None
            url_name = self.format_url_pattern(url_pattern)
            view = url_pattern.callback
            provided_view_kwargs = url_pattern.default_args
            # "kwargs" argument provided to path()
            if isinstance(url_pattern.pattern, RoutePattern):
                arg_names = list(url_pattern.pattern.converters.keys())
            else:
                arg_names = list(url_pattern.pattern.regex.groupindex)
            reverse_kwargs = {k: common_kwargs[k] for k in arg_names if k in common_kwargs}
            for user in self.get_users():
                username = None if user.is_anonymous else user.username
                request_testers_: ExpectedResponses = expected_responses.get(url_name, {}).get(username, Http404)
                if not isinstance(request_testers_, list):
                    request_testers: List[ExpectedResponse] = [request_testers_]
                else:
                    request_testers: List[ExpectedResponse] = request_testers_
                for request_tester in request_testers:
                    if not isinstance(request_tester, RequestTester):
                        request_tester = RequestTester(request_tester)
                    # actual view kwargs are the kwargs provided in the get_urls()
                    self.__class__.views_test_counter += 1
                    exc = request_tester.evaluate(
                        self,
                        view_name,
                        view,
                        reverse_kwargs,
                        provided_view_kwargs,
                        user,
                    )
                    if exc is not None:
                        failed_url = request_tester.request.path + (
                            "?" + iri_to_uri(request_tester.request.META.get("QUERY_STRING", ""))
                            if request_tester.request.META.get("QUERY_STRING", "")
                            else ""
                        )
                        failure = (
                            url_pattern,
                            username,
                            exc,
                            failed_url,
                        )
                        failures.append(failure)
        for url_pattern, username, e, url in failures:
            pattern = self.format_url_pattern(url_pattern)
            if username is None:
                username = "AnonymousUser"
            else:
                username = f"user {username}"
            base_url = settings.SERVER_BASE_URL[:-1]
            msg = f"{display_prefix}Exception in {pattern} with {username}: {e} ({base_url}{url})."
            print(msg)
        self.assertEqual(0, len(failures))

    @staticmethod
    def format_url_pattern(url_pattern) -> str:
        """Return a string representation of the provided URL pattern."""
        if url_pattern.name:
            return url_pattern.name
        if isinstance(url_pattern.pattern, RoutePattern):
            return url_pattern.pattern._route
        return url_pattern.pattern._regex


class TestModelAdmin(TestMultipleViews):
    """Define some methods for testing *all* views defined in the get_urls() of a ModelAdmin.

     Each view is tested with different users.
    . A RequestTester define more precisely how to test a given view with a given user.

    Many of these views require some kwargs for the `reverse` function, but most of them have always the same meaning.
    So, we always reuse the same kwargs for all views (but we can override them in a RequestTester).

    `expected_responses` is a dict such that:
    - the key is the view name
    - the value is a dict such that:
        - the key is a username (or None for AnonymousUser)
        - the value is any of:
            - an integer (the expected HTTP status code — then this is the only checked thing)
            - an exception (only the type of the raised exception is checked)
            - a HttpResponse (only status_codes are checked)
            - a callable that will be given the TestCase and the HttpResponse as arguments
            - a RequestTester to deeply customize the test (including the request args/kwargs)
            - a list of any of the previous values

    """

    checked_model_admins = set()
    placeholders = "%(app_label)s_%(model_name)s_"
    expected_responses: Dict[str, Dict[Optional[str], ExpectedResponses]] = {
        f"{placeholders}changelist": {
            None: 302,
            "staff": PermissionDenied,
            "admin": 200,
        },
        f"{placeholders}add": {
            None: 302,
            "staff": PermissionDenied,
            "admin": 200,
        },
        f"{placeholders}history": {
            None: 302,
            "staff": PermissionDenied,
            "admin": 200,
        },
        f"{placeholders}delete": {
            None: 302,
            "staff": PermissionDenied,
            "admin": 200,
        },
        f"{placeholders}change": {
            None: 302,
            "staff": PermissionDenied,
            "admin": 200,
        },
        "<path:object_id>/": {
            None: 302,
            "staff": 302,
            "admin": 302,
        },
    }

    # expected_responses["view_name"]["staff"] = 200
    # %(app_label)s and %(model_name)s can
    # be used in the view names to avoid repeating the same values.
    def check_model_admin(
        self,
        model_admin: Union[Type[admin.ModelAdmin], Type[models.Model]],
        common_kwargs: Dict[str, str],
        display_prefix: str = "",
        expected_responses: Dict[str, Dict[Optional[str], ExpectedResponses]] = None,
    ):
        """Check all admin views.

        :param model_admin: the models.Model or admin.ModelAdmin to check
        :param common_kwargs: dict of values for reversing URLs
        :param display_prefix: prefix to display in front of each error
        :param expected_responses: extra responses, only for this ModelAdmin
            (same format as self.expected_responses)
        :return:
        """
        if isinstance(model_admin, type) and issubclass(model_admin, models.Model):
            model_admin = default_site.get_model_admin(model_admin)

        TestModelAdmin.checked_model_admins.add(model_admin)
        base_expected_responses = self.get_expected_responses(model_admin, raw_expected_responses=expected_responses)
        # with app_label="app", model_name="model", if we have both
        #   "%(app_label)s_%(model_name)s_viewname" and "app_model_viewname" keys  in raw_expected_responses
        #   we do not want to override the initial values.
        url_patterns = self.get_admin_urls(model_admin)
        self.check_url_patterns(
            url_patterns,
            common_kwargs,
            base_expected_responses,
            display_prefix=display_prefix,
        )

    def get_expected_responses(
        self,
        model_admin: admin.ModelAdmin,
        raw_expected_responses: Dict[str, Dict[Optional[str], ExpectedResponses]] = None,
    ) -> Dict[str, Dict[Optional[str], ExpectedResponses]]:
        """Return the expected responses for this model_admin.

        Replace %(app_label)s and %(model_name)s placeholders in the view names and use
        them when these view names are missing from the expected_responses.
        """
        base_expected_responses = {}
        base_expected_responses.update(self.expected_responses)
        if raw_expected_responses is not None:
            base_expected_responses.update(raw_expected_responses)

        opts = {
            "app_label": model_admin.model._meta.app_label,
            "model_name": model_admin.model._meta.model_name,
        }
        expected_responses = {url_name % opts: v for (url_name, v) in base_expected_responses.items() if url_name}
        expected_responses.update(base_expected_responses)
        return expected_responses

    # noinspection PyMethodMayBeStatic
    def get_admin_urls(self, model_admin: admin.ModelAdmin) -> List[Tuple[str, URLPattern]]:
        """Return the list of all URLs in this model_admin."""
        return [("admin:", x) for x in model_admin.get_urls()]


class TestModel(TestModelAdmin):
    """Define some methods for testing all views defined in the get_urls() of a ModelAdmin."""

    def get_object(self) -> models.Model:
        """Return an object of the model to test."""
        raise NotImplementedError("You must override this method to return an object of the model to test.")

    def test_model_admin(self):
        """Test all views of the model admin correspoding to the provided object."""
        try:
            obj: models.Model = self.get_object()
        except NotImplementedError as e:
            self.skipTest(str(e))
        if obj._state.adding:
            obj.save()
        common_kwargs = self.get_common_kwargs_for_object(obj)
        self.check_model_admin(obj.__class__, common_kwargs=common_kwargs, display_prefix="")

    def get_common_kwargs_for_object(self, obj) -> Dict[str, str]:
        """Return the common kwargs for all views of the provided object."""
        return {"object_id": str(obj.pk)}



"""List of URLs for the df_site app."""

from django.conf import settings
from django.urls import path
from django.utils.module_loading import import_string

from df_site.users.views import theme_switch

app_name = "users"
urlpatterns = [
    path("theme-switch", theme_switch, name="theme_switch"),
]
if settings.AUTH_USER_SETTINGS_VIEW:
    user_settings_view = import_string(settings.AUTH_USER_SETTINGS_VIEW)
    urlpatterns.append(
        path("settings", user_settings_view.as_view(), name="settings"),
    )



"""Django settings for the project."""

from typing import List, Tuple

from df_config.config.dynamic_settings import CallableSetting, SettingReference
from django.utils.translation import gettext_lazy as _

from df_site.dynamic_settings import allauth_signup_form, are_tests_running, load_tox_environment

load_tox_environment()
DF_SITE_TITLE = "More batteries to Django"
DF_SITE_SECURITY_EMAIL = SettingReference("ADMIN_EMAIL")
DF_SITE_SECURITY_LANGUAGE_CODE = SettingReference("LANGUAGE_CODE")
DF_SITE_SECURITY_GPG_CONTENT = None

DF_SITE_DESCRIPTION = "A simple Django site with lots of batteries included."
DF_SITE_KEYWORDS = ["Django", "Bootstrap", "WebSockets", "HTMX", "Django Channels"]
DF_SITE_AUTHOR = "d9pouces"
DF_SITE_ORGANIZATION = "d9pouces"
DF_SITE_SOCIAL_NETWORKS = {
    "instagram": "https://www.instagram.com/d9pouces/",
    "twitter": "https://x.com/d9pouces/",
    "github": "https://github.com/d9pouces/",
}

DF_SITE_THEMES: List[Tuple[str, str, str]] = [  # ('theme name', 'theme label', 'icon name')
    ("auto", _("Auto"), "toggle-on"),
    ("light", _("Light"), "sun"),
    ("dark", _("Dark"), "moon"),
]
DF_ANDROID_THEME_COLOR = "#ffffff"
DF_ANDROID_BACKGROUND_COLOR = "#ffffff"
DF_MICROSOFT_BACKGROUND_COLOR = "#da532c"
DF_SAFARI_PINNED_COLOR = "#5bbad5"

DF_IMAGE_THUMBNAILS = {
    "media": {
        "src_storage": "default",
        "dst_storage": "staticfiles",
        "cache": "default",
        "prefix": "T",
        "reversible": True,
    },
    "static": {
        "src_storage": "staticfiles",
        "dst_storage": "staticfiles",
        "cache": "default",
        "prefix": "S",
        "reversible": True,
    },
}

CSP_IMG_SRC = ["'self'", "data: w3.org/svg/2000", "https://log.pinterest.com"]
# "https://log.pinterest.com" is used for Pinterest
CSP_STYLE_SRC = ["'self'"]
CSP_FONT_SRC = ["'self'"]
CSP_DEFAULT_SRC = ["'none'"]
CSP_SCRIPT_SRC = [
    "'self'",
    "https://www.google.com",
    "https://www.gstatic.com",
    "https://assets.pinterest.com",
    "'unsafe-inline'",
]
# unsafe-inline is used for django-allauth (inline JS)
# "https://www.google.com", "https://www.gstatic.com" are used for reCAPTCHA
# "https://assets.pinterest.com" is used for Pinterest
CSP_OBJECT_SRC = ["'self'"]
CSP_MEDIA_SRC = ["'self'"]
CSP_FRAME_SRC = ["'self'", "https://www.google.com", "https://assets.pinterest.com"]
# "https://www.google.com" is used for reCAPTCHA
# "https://assets.pinterest.com" is used for Pinterest
CSP_CHILD_SRC = ["'self'"]
CSP_FRAME_ANCESTORS = ["'self'"]
CSP_FORM_ACTION = ["'self'"]
CSP_MANIFEST_SRC = ["'self'"]
CSP_BASE_URI = ["'self'"]
CSP_REPORT_URI = "/csp-report/"
CSP_REPORT_TO = None
DF_TEMPLATE_CONTEXT_PROCESSORS = [
    "df_site.context_processors.global_site_infos",
    "django.template.context_processors.request",
]
TESTING = CallableSetting(are_tests_running)
RECAPTCHA_PUBLIC_KEY = ""
RECAPTCHA_PRIVATE_KEY = ""
DF_INSTALLED_APPS = [
    "django_bootstrap5",
    "df_site.apps.DFSiteApp",
    "cookie_consent",
    "postman",
    "allauth.mfa",
    "allauth.usersessions",
    "django_ckeditor_5",
    "django_recaptcha",
]

DF_MIDDLEWARE = [
    "df_site.middleware.websocket_middleware",
]

USERSESSIONS_TRACK_ACTIVITY = True
POSTMAN_DISALLOW_ANONYMOUS = True
POSTMAN_AUTO_MODERATE_AS = True
POSTMAN_I18N_URLS = False
AUTH_USER_MODEL = "df_site.PreferencesUser"
AUTH_USER_SETTINGS_VIEW = "df_site.users.views.UserSettingsView"
COOKIE_CONSENT_SECURE = SettingReference("USE_SSL")
COOKIE_CONSENT_DOMAIN = "{SERVER_NAME}"
COOKIE_CONSENT_SAMESITE = "Strict"
PIPELINE = {
    "PIPELINE_ENABLED": SettingReference("PIPELINE_ENABLED"),
    "JAVASCRIPT": {
        "base": {
            "source_filenames": [
                "js/base.js",
                "js/df_websockets.min.js",
                # "django_ckeditor_5/dist/bundle.js"
            ],
            "output_filename": "js/base.min.js",
            #            "integrity": "sha384",
            "crossorigin": "anonymous",
            "extra_context": {
                "defer": True,
            },
        },
        "app": {
            "source_filenames": ["js/app.ts"],
            "output_filename": "js/app.min.js",
            #            "integrity": "sha384",
            "crossorigin": "anonymous",
            "extra_context": {
                "defer": True,
            },
        },
    },
    "STYLESHEETS": {
        "base": {
            "source_filenames": [
                "django_ckeditor_5/src/override-django.css",
                "css/ckeditor5.css",
                "css/base.css",
            ],
            "output_filename": "css/base.min.css",
            "extra_context": {"media": "all"},
            #            "integrity": "sha384",
            "crossorigin": "anonymous",
        },
        "app": {
            "source_filenames": ["css/app.css"],
            "output_filename": "css/app.min.css",
            "extra_context": {"media": "all"},
            #            "integrity": "sha384",
            "crossorigin": "anonymous",
        },
    },
    "CSS_COMPRESSOR": SettingReference("PIPELINE_CSS_COMPRESSOR"),
    "JS_COMPRESSOR": SettingReference("PIPELINE_JS_COMPRESSOR"),
    "COMPILERS": SettingReference("PIPELINE_COMPILERS"),
}
SOCIALACCOUNT_EMAIL_AUTHENTICATION_AUTO_CONNECT = True
SOCIALACCOUNT_LOGIN_ON_GET = True
MFA_SUPPORTED_TYPES = ["totp", "webauthn", "recovery_codes"]
MFA_PASSKEY_LOGIN_ENABLED = True
MFA_WEBAUTHN_ALLOW_INSECURE_ORIGIN = SettingReference("DEBUG")
ACCOUNT_LOGIN_BY_CODE_ENABLED = True
ACCOUNT_EMAIL_VERIFICATION = "optional"
ACCOUNT_SIGNUP_FORM_CLASS = CallableSetting(allauth_signup_form)

CKEDITOR_5_USER_LANGUAGE = True
special_chars = [
    {"title": _("smiley face"), "character": "😊"},
    {"title": _("smiley face"), "character": ":)"},
]
CK_EDITOR_5_UPLOAD_FILE_VIEW = "django_ckeditor_5.views.upload_file"
CK_EDITOR_5_UPLOAD_FILE_VIEW_NAME = "upload_file"
CKEDITOR_5_CONFIGS = {
    "inline": {
        "specialChars": special_chars,
        "toolbar": [
            "bold",
            "italic",
            "underline",
            "strikethrough",
            "subscript",
            "superscript",
            "|",
            "specialCharacters",
            "removeFormat",
        ],
        "plugins": [
            "Essentials",
            "Autoformat",
            "Bold",
            "Italic",
            "Underline",
            "Strikethrough",
            "Code",
            "Subscript",
            "Superscript",
            "Paragraph",
            "Font",
            "PasteFromOffice",
            "RemoveFormat",
            "Highlight",
            "SpecialCharacters",
            "SpecialCharactersEssentials",
            "ShowBlocks",
            "SelectAll",
        ],
    },
    "inline_link": {
        "specialChars": special_chars,
        "toolbar": [
            "bold",
            "italic",
            "underline",
            "strikethrough",
            "subscript",
            "superscript",
            "link",
            "|",
            "specialCharacters",
            "removeFormat",
        ],
        "plugins": [
            "Essentials",
            "Autoformat",
            "Bold",
            "Italic",
            "Underline",
            "Strikethrough",
            "Code",
            "Subscript",
            "Superscript",
            "Link",
            "Paragraph",
            "Font",
            "PasteFromOffice",
            "RemoveFormat",
            "Highlight",
            "SpecialCharacters",
            "SpecialCharactersEssentials",
            "ShowBlocks",
            "SelectAll",
        ],
    },
    "default": {
        "specialChars": special_chars,
        "toolbar": [
            "heading",
            "bulletedList",
            "numberedList",
            "blockQuote",
            "|",
            "bold",
            "italic",
            "underline",
            "strikethrough",
            "subscript",
            "superscript",
            "link",
            "highlight",
            "|",
            "insertTable",
            "insertImage",
            "specialCharacters",
            "removeFormat",
            "undo",
            "redo",
        ],
        "image": {
            "toolbar": [
                "imageTextAlternative",
                "|",
                "imageStyle:full",
                "imageStyle:alignLeft",
                "imageStyle:alignRight",
                "imageStyle:alignCenter",
                "imageStyle:side",
                "|",
            ],
            "styles": [
                "full",
                "side",
                "alignLeft",
                "alignRight",
                "alignCenter",
            ],
        },
        "list": {"properties": {"styles": False, "startIndex": True, "reversed": True}},
        "table": {
            "defaultHeadings": {"rows": 1, "columns": 1},
            "contentToolbar": ["tableColumn", "tableRow", "mergeTableCells"],
        },
    },
}



"""Customized postman views."""

from django.contrib.auth.base_user import AbstractBaseUser
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.views.generic import View
from postman.models import Message
from postman.views import UpdateDualMixin


class MessageUpdateView(UpdateDualMixin, View):
    """View for updating messages or conversations.

    This single view replaces the following views: `ArchiveView`, `DeleteView`,
    `UndeleteView`, 'MarkReadView', 'MarkUnreadView'.
    This is required to avoid the onclick attribute in the template (forbidden by CSP).
    """

    @property
    def selected_action(self):
        """Get the action to perform on the selected messages or conversations."""
        action = self.request.POST.get("action")
        if action not in {"delete", "archive", "undelete", "read", "unread"}:
            raise PermissionDenied
        return action

    def _action(self, user: AbstractBaseUser, filter_: Q):
        """Perform the action on the selected messages or conversations."""
        if self.selected_action in {"read", "unread"}:
            Message.objects.as_recipient(user, filter_).filter(
                **{f"{self.field_bit}__isnull": bool(self.field_value)}
            ).update(**{self.field_bit: self.field_value})
        else:
            super()._action(user, filter_)
        # an empty set cannot be estimated as an error, it may be just a badly chosen selection

    @property
    def field_bit(self):
        """Get the field to update."""
        return {
            "delete": "deleted_at",
            "archive": "archived",
            "undelete": "deleted_at",
            "read": "read_at",
            "unread": "read_at",
        }[self.selected_action]

    @property
    def success_msg(self):
        """Get the success message to display."""
        return {
            "delete": _("Messages or conversations successfully deleted."),
            "archive": _("Messages or conversations successfully archived."),
            "undelete": _("Messages or conversations successfully recovered."),
            "read": _("Messages or conversations successfully marked as read."),
            "unread": _("Messages or conversations successfully marked as unread."),
        }[self.selected_action]

    @property
    def field_value(self):
        """Get the value to set the field to."""
        n = now()
        return {"delete": n, "archive": True, "undelete": None, "read": n, "unread": None}[self.selected_action]



"""Override default admin filters with a custom template."""

from django.contrib.admin import AllValuesFieldListFilter as AllValuesFieldListFilterBase
from django.contrib.admin import BooleanFieldListFilter as BooleanFieldListFilterBase
from django.contrib.admin import ChoicesFieldListFilter as ChoicesFieldListFilterBase
from django.contrib.admin import DateFieldListFilter as DateFieldListFilterBase
from django.contrib.admin import EmptyFieldListFilter as EmptyFieldListFilterBase
from django.contrib.admin import FieldListFilter as FieldListFilterBase
from django.contrib.admin import RelatedFieldListFilter as RelatedFieldListFilterBase
from django.contrib.admin import RelatedOnlyFieldListFilter as RelatedOnlyFieldListFilterBase
from django.contrib.admin import SimpleListFilter as SimpleListFilterBase


class ListFilterMixin:
    """A mixin for list filters, replacing the default template."""

    template = "df_components/list_filter.html"


class SimpleListFilter(ListFilterMixin, SimpleListFilterBase):
    """A filter for simple fields."""

    pass


class FieldListFilter(ListFilterMixin, FieldListFilterBase):
    """A filter for list fields."""

    pass


class RelatedFieldListFilter(ListFilterMixin, RelatedFieldListFilterBase):
    """A filter for related fields."""

    def field_admin_ordering(self, field, request, model_admin):
        """Return the default ordering for related field, skipping the use of the admin site."""
        # noinspection PyProtectedMember
        return field.remote_field.model._meta.ordering


class BooleanFieldListFilter(ListFilterMixin, BooleanFieldListFilterBase):
    """A filter for boolean fields."""

    pass


class ChoicesFieldListFilter(ListFilterMixin, ChoicesFieldListFilterBase):
    """A filter for choice fields."""

    pass


class DateFieldListFilter(ListFilterMixin, DateFieldListFilterBase):
    """A filter for date fields."""

    pass


class AllValuesFieldListFilter(ListFilterMixin, AllValuesFieldListFilterBase):
    """A filter for all values of a field."""

    pass


class RelatedOnlyFieldListFilter(ListFilterMixin, RelatedOnlyFieldListFilterBase):
    """A filter for related fields."""

    pass


class EmptyFieldListFilter(ListFilterMixin, EmptyFieldListFilterBase):
    """A filter for empty fields."""

    pass



"""Python functions to provide Django settings depending on other settings."""

import os
from typing import Any, Optional
from urllib.parse import urlparse


def allauth_signup_form(values: dict[str, Any]) -> Optional[str]:
    """Return the form class to use for signing up."""
    if values.get("RECAPTCHA_PRIVATE_KEY") and values.get("RECAPTCHA_PUBLIC_KEY"):
        return "df_site.users.forms.ReCaptchaForm"
    return None


allauth_signup_form.required_settings = ["RECAPTCHA_PRIVATE_KEY", "RECAPTCHA_PUBLIC_KEY"]


def are_tests_running(values: dict[str, Any]) -> bool:
    """Return True if we are running unit tests."""
    return "testserver" in values.get("ALLOWED_HOSTS", [])


are_tests_running.required_settings = ["ALLOWED_HOSTS"]


def patch_tox_environment(url_variable: str, host_variable: str, port_variable: str):
    """Patch a single URL for the workaround for https://github.com/tox-dev/tox-docker/issues/55."""
    if not all(x in os.environ for x in (host_variable, port_variable, url_variable)):
        return
    parsed_url = urlparse(os.environ[url_variable])
    if parsed_url.username or parsed_url.password:
        netloc = f"{parsed_url.username}:{parsed_url.password}@{os.environ[host_variable]}:{os.environ[port_variable]}"
    else:
        netloc = f"{os.environ[host_variable]}:{os.environ[port_variable]}"
    os.environ[url_variable] = parsed_url._replace(netloc=netloc).geturl()


def load_tox_environment():
    """Is a workaround for https://github.com/tox-dev/tox-docker/issues/55."""
    patch_tox_environment("REDIS_URL", "REDIS_HOST", "REDIS_6379_TCP_PORT")
    patch_tox_environment("DATABASE_URL", "POSTGRES_HOST", "POSTGRES_5432_TCP_PORT")
    patch_tox_environment("MAIN_STORAGE_DIR", "MINIO_HOST", "MINIO_9000_TCP_PORT")



"""This module contains the functions related to the notifications of the users."""

from typing import Optional, Union

from django.contrib.auth.models import AbstractUser
from django.contrib.sites.models import Site


# noinspection PyUnusedLocal
def email_address_on_message(user: Optional[AbstractUser], action: str, site: Site) -> Union[bool, str]:
    """Return the email address of the user if it is authenticated and if its accepts emails."""
    if user and user.is_authenticated and user.email_notifications:
        return user.email
    return False



"""Configuration for the df_site app."""

from typing import Type

from django.apps import AppConfig
from django.db.models.signals import post_migrate, pre_save
from django.utils.translation import gettext_lazy as _


class DFSiteApp(AppConfig):
    """Configuration for the df_site app."""

    default_auto_field = "django.db.models.AutoField"
    name = "df_site"
    verbose_name = _("df_site")

    def ready(self):
        """Run code when the app is ready."""
        super().ready()
        post_migrate.connect(auto_create_site_object, sender=self)
        from django.contrib.auth import get_user_model

        user_model = get_user_model()
        pre_save.connect(create_admin_user, sender=user_model)


def create_admin_user(sender, instance, **kwargs):
    """Create an admin user if none already exists."""
    from django.contrib.auth.models import AbstractUser

    sender: Type[AbstractUser]
    instance: AbstractUser
    if hasattr(create_admin_user, "run"):
        return
    if not hasattr(instance, "is_superuser") or sender.objects.filter(is_superuser=True, is_active=True).exists():
        return
    instance.is_superuser = True
    instance.is_staff = True
    setattr(create_admin_user, "run", True)


# noinspection PyUnusedLocal
def auto_create_site_object(sender, **kwargs):
    """Create a Site object if it does not exist."""
    from django.conf import settings
    from django.contrib.sites.models import Site

    site, created = Site.objects.get_or_create(
        id=1,
        defaults={
            "domain": settings.SERVER_NAME,
            "name": settings.DF_SITE_TITLE,
        },
    )
    site.domain = settings.SERVER_NAME
    site.name = settings.DF_SITE_TITLE
    site.save()



"""Contains constants for the users app."""

import re

BRAND_ICONS = {
    "amazon": "amazon",
    "amazon_cognito": "amazon",
    "angellist": "angellist",
    "apple": "apple",
    "atlassian": "atlassian",
    "battlenet": "battle-net",
    "bitbucket_oauth2": "bitbucket",
    "digitalocean": "digital-ocean",
    "discord": "discord",
    "dropbox": "dropbox",
    "evernote": "evernote",
    "facebook": "facebook",
    "figma": "figma",
    "fivehundredpx": "500px",
    "flickr": "flickr",
    "foursquare": "foursquare",
    "github": "github",
    "gitlab": "gitlab",
    "google": "google",
    "hubspot": "hubspot",
    "instagram": "instagram",
    "line": "line",
    "linkedin_oauth2": "linkedin",
    "mailchimp": "mailchimp",
    "mailru": "",
    "meetup": "meetup",
    "microsoft": "microsoft",
    "odnoklassniki": "odnoklassniki",
    "openid": "openid",
    "openid_connect": "openid",
    "orcid": "orcid",
    "patreon": "patreon",
    "paypal": "paypal",
    "pinterest": "pinterest",
    "pocket": "get-pocket",
    "reddit": "reddit",
    "salesforce": "salesforce",
    "shopify": "shopify",
    "slack": "slack",
    "snapchat": "snapchat",
    "soundcloud": "soundcloud",
    "spotify": "spotify",
    "stackexchange": "stack-exchange",
    "steam": "steam",
    "strava": "strava",
    "stripe": "stripe",
    "telegram": "telegram",
    "tiktok": "tiktok",
    "trello": "trello",
    "tumblr": "tumblr",
    "twitch": "twitch",
    "twitter": "twitter",
    "twitter_oauth2": "twitter",
    "untappd": "untappd",
    "vimeo": "vimeo",
    "vimeo_oauth2": "vimeo",
    "vk": "vk",
    "weibo": "weibo",
    "weixin": "weixin",
    "windowslive": "windows",
    "xing": "xing",
    "yahoo": "yahoo",
    "yandex": "yandex",
}
INT_RE = re.compile(r"^(0|[1-9]\d*)$")
SIZE_RE = re.compile(r"^([1-9]\d*)x([1-9]\d*)$")



"""Forms for the Postman application."""

from django import forms
from django.utils.translation import gettext_lazy as _
from postman.forms import AnonymousWriteForm, FullReplyForm, QuickReplyForm, WriteForm

from df_site.form_fields import CKEditor5Field


class CKEditor5Form(forms.Form):
    """Form with a CKEditor 5 widget."""

    body = CKEditor5Field(label=_("body"), required=True, config_name="default")
    subject = CKEditor5Field(label=_("subject"), required=True, config_name="inline")


class HTMLWriteForm(CKEditor5Form, WriteForm):
    """Form used by authenticated users."""

    pass


class HTMLAnonymousWriteForm(CKEditor5Form, AnonymousWriteForm):
    """Form used by anonymous users."""

    pass


class HTMLFullReplyForm(CKEditor5Form, FullReplyForm):
    """Form for complete Postman replies."""

    pass


class HTMLQuickReplyForm(CKEditor5Form, QuickReplyForm):
    """Form for postman quick replies."""

    pass



"""Define the context processors with global variables about the site."""

import datetime
import logging
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional

from django.conf import settings
from django.db.models import Q
from django.http import HttpRequest
from django.utils.timezone import now

from df_site.models import AlertRibbon
from df_site.user_settings import get_user_setting

logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def get_current_ribbons(current_date: datetime.datetime) -> Iterable[AlertRibbon]:
    """Get the active ribbons for the current date.

    The current date will be rounded to the nearest hour.
    """
    return list(
        AlertRibbon.objects.filter(
            Q(Q(end_date__gte=current_date) & Q(end_date__isnull=False)) | Q(end_date__isnull=True),
            Q(Q(start_date__lte=current_date) & Q(start_date__isnull=False)) | Q(start_date__isnull=True),
            is_active=True,
        )
    )


def global_site_infos(request: HttpRequest, current_date: Optional[datetime.datetime] = None) -> Dict[str, Any]:
    """Adds a few values to the request context."""
    try:
        color_theme = get_user_setting("color_theme", request=request)
    except Exception as e:
        logger.error("Error getting color theme: %s", e)
        color_theme = settings.DF_SITE_THEMES[0][0]
    if current_date is None:
        current_date = now()
    current_date = current_date.replace(minute=0, second=0, microsecond=0)
    ribbons = get_current_ribbons(current_date)
    x_url = settings.DF_SITE_SOCIAL_NETWORKS.get("twitter", "")
    __, __, df_site_x_account = x_url.rpartition("/")
    return {
        "DF_SITE_TITLE": settings.DF_SITE_TITLE,
        "DF_SITE_DESCRIPTION": settings.DF_SITE_DESCRIPTION,
        "DF_SITE_KEYWORDS": settings.DF_SITE_KEYWORDS,
        "DF_SITE_AUTHOR": settings.DF_SITE_AUTHOR,
        "DF_SITE_ORGANIZATION": settings.DF_SITE_ORGANIZATION,
        "DF_SITE_X_ACCOUNT": df_site_x_account,
        "DF_SITE_SOCIAL_NETWORKS": settings.DF_SITE_SOCIAL_NETWORKS.items(),
        "DF_COLOR_THEMES": settings.DF_SITE_THEMES,
        "DF_MICROSOFT_BACKGROUND_COLOR": settings.DF_MICROSOFT_BACKGROUND_COLOR,
        "DF_ANDROID_THEME_COLOR": settings.DF_ANDROID_THEME_COLOR,
        "DF_SAFARI_PINNED_COLOR": settings.DF_SAFARI_PINNED_COLOR,
        "COLOR_THEME": color_theme,
        "DF_RIBBONS": ribbons,
    }



"""Template tags for image generation in cache."""

import datetime
import hashlib
import logging
import math
import os
import re
import time
from enum import IntEnum
from html import escape
from typing import Callable, Dict, Iterable, List, Optional, Set, Union

from django import template
from django.conf import settings
from django.core.cache import BaseCache, caches
from django.core.files.storage import Storage, storages
from django.core.files.storage.filesystem import FileSystemStorage
from django.db.models.fields.files import FieldFile
from django.http import Http404, HttpRequest
from django.urls import reverse
from django.utils.safestring import mark_safe
from PIL import Image

register = template.Library()
logger = logging.getLogger(__name__)


class AspectPolicy(IntEnum):
    """Aspect aspect_policy of resized images."""

    FORCE_HEIGHT_CROP = 0  # height will be respected, width will be cropped or too small
    FORCE_WIDTH_CROP = 1  # width will be respected, height will be cropped or too small
    FORCE_MIN_CROP = 2  # the smallest dimension will be respected, the other will be cropped
    FORCE_MAX_FIT = 3  # the largest dimension will be respected, the other will be too small
    FORCE_MAX_TRANSPARENT = 4  # the largest dimension will be respected, the other extra will be transparent

    @classmethod
    def get_policy(cls, int_policy: int) -> "AspectPolicy":
        """Return the AspectPolicy corresponding to the given integer."""
        for policy in AspectPolicy:
            if policy.value == int_policy:
                return policy
        raise ValueError(f"Invalid aspect policy: {int_policy}")


@register.simple_tag(takes_context=True)
def media_image(
    context,
    path: Union[str, FieldFile, callable],
    widths: Optional[Union[str, List[int]]] = None,
    write: bool = True,
    aspect_policy: AspectPolicy = AspectPolicy.FORCE_MAX_TRANSPARENT,
    fmt: str = "webp",
    height: Optional[int] = None,
    width: Optional[int] = None,
    alt: str = "",
    crossorigin: str = "anonymous",
    loading: str = "lazy",
    referrerpolicy: str = "same-origin",
    **kwargs,
):
    """Generate a <img> tag with associated attributes for a media file.

    :param context: the template context
    :param path: the path to the image
    :param storage: where thumbnail images will be stored
    :param widths: list of widths to create, given in the src-set attribute
    :param write: whether to actually create the thumbnail images
    :param aspect_policy: the policy to apply about the aspect ratio
    :param fmt: the format of the thumbnail images
    :param height: the targeted height of the image
    :param width: the targeted width of the image
    :param alt: the alt attribute, passed as-is to the <img> tag
    :param crossorigin: the crossorigin attribute, passed as-is to the <img> tag
    :param loading: the loading attribute, passed as-is to the <img> tag
    :param referrerpolicy: the referrerpolicy attribute, passed as-is to the <img> tag
    :param kwargs: additional attributes that are passed as-is to the <img> tag
    """
    if isinstance(widths, str):
        widths = [int(x.strip()) for x in widths.split(",")]
    if callable(path):
        path = path()
    elif isinstance(path, FieldFile):
        path: str = path.name
    img = CachedImage(
        CachedImage.sources["media"]["src_storage"],
        CachedImage.sources["media"]["dst_storage"],
        CachedImage.sources["media"]["cache"],
        path,
        CachedImage.sources["media"]["prefix"],
        height=height,
        width=width,
        widths=widths,
        aspect_policy=aspect_policy,
        fmt=fmt,
        write_thumbnails=write,
    )
    img.process()
    result = img.as_html_tag(alt=alt, crossorigin=crossorigin, loading=loading, referrerpolicy=referrerpolicy, **kwargs)
    if result == "":
        img.log_error(context.get("request"))
    return result


@register.simple_tag(takes_context=True)
def static_image(
    context,
    path: Union[callable, str],
    widths: Optional[Union[str, List[int]]] = None,
    write: bool = True,
    aspect_policy: AspectPolicy = AspectPolicy.FORCE_MAX_TRANSPARENT,
    fmt: str = "webp",
    height: Optional[int] = None,
    width: Optional[int] = None,
    alt: str = "",
    crossorigin: str = "anonymous",
    loading: str = "lazy",
    referrerpolicy: str = "same-origin",
    **kwargs,
):
    """Generate a <img> tag with associated attributes for a static file."""
    if callable(path):
        path = path()
    if isinstance(widths, str):
        widths = [int(x.strip()) for x in widths.split(",")]
    img = CachedImage(
        CachedImage.sources["static"]["src_storage"],
        CachedImage.sources["static"]["dst_storage"],
        CachedImage.sources["static"]["cache"],
        path,
        CachedImage.sources["static"]["prefix"],
        height=height,
        width=width,
        widths=widths,
        aspect_policy=aspect_policy,
        fmt=fmt,
        write_thumbnails=write,
    )
    img.process()
    result = img.as_html_tag(alt=alt, crossorigin=crossorigin, loading=loading, referrerpolicy=referrerpolicy, **kwargs)
    if result == "":
        img.log_error(context.get("request"))
    return result


IMAGE_THUMBNAILS = getattr(
    settings,
    "DF_IMAGE_THUMBNAILS",
    {
        "media": {
            "src_storage": "default",
            "dst_storage": "staticfiles",
            "cache": "default",
            "prefix": "T",
            "reversible": True,
        },
        "static": {
            "src_storage": "staticfiles",
            "dst_storage": "staticfiles",
            "cache": "default",
            "prefix": "S",
            "reversible": True,
        },
    },
)


class CachedImage:
    """Handle original images to create required thumbnails."""

    src_fast_check_expiration = 86400
    src_slow_check_expiration = 86400 * 30
    src_cache_expiration = 86400 * 365
    default_cache_data = {
        "width": None,
        "height": None,
        "filesize": None,
        "sha256": None,
        "next_fast_check": None,
        "next_slow_check": None,
        "mtime": None,
    }
    default_widths = [64, 160, 320, 640, 1280, 1920, 3840]
    formats = ["webp", "jpg", "jpeg", "png"]
    target_path_re = re.compile(
        r"(?P<cache_prefix>[^/]+)/"
        r"(?P<width>\d+)x(?P<height>\d+)_(?P<aspect_policy>[^/]+)/"
        r"(?P<src_path>.+)\.(?P<fmt>[^.]+)$",
    )
    sources: Dict[str, Dict[str, str]] = IMAGE_THUMBNAILS

    def __init__(
        self,
        src_storage: str,
        dst_storage: str,
        cache: str,
        src_path: str,
        cache_prefix: str,
        url: Callable[[str], str] = lambda p: reverse("thumbnails", kwargs={"path": p}),
        height: Optional[int] = None,
        width: Optional[int] = None,
        widths: Optional[List[int]] = None,
        aspect_policy: AspectPolicy = AspectPolicy.FORCE_MAX_TRANSPARENT,
        fmt: str = "webp",
        write_thumbnails: bool = True,
    ):
        """Initialize the CachedImage."""
        dst_name = hashlib.sha256(src_path.encode()).hexdigest()
        cache_key = f"{cache_prefix}{dst_name}"

        self.src_storage_obj: Storage = storages[src_storage]
        self.dst_storage_obj: Storage = storages[dst_storage]
        self.cache_obj: BaseCache = caches[cache]
        self.src_path = src_path
        self.cache_key = cache_key
        self.cache_prefix = cache_prefix
        self.url: Callable[[str], str] = url  # function to generate the URL of the thumbnails from the thumbnail path
        self.target_height = height  # target height of the image
        self.target_width = width  # target width of the image
        self.widths = widths  # list of widths to create if possible
        self.aspect_policy = aspect_policy  # aspect ratio policy when target width and height are given
        self.fmt = fmt  # format ('webp', 'jpg', 'jpeg', 'png') of the thumbnails
        self.cache_data: Dict[str, Union[None, str, int, Dict]] = self.default_cache_data.copy()
        self.cache_changed: bool = False  # whether the cache has been changed and must be saved
        self.write_thumbnail: bool = write_thumbnails  # whether the thumbnails must be actually created
        self.paths_srcset: Dict[str, str] = {}  # the sizes and paths for the srcset attribute
        self.path_src: str = ""  # the path used in the src attribute
        self.created_sizes: Set[int] = set()  # the list of actually created sizes

    def process(self):
        """Process the image, updating the cache if necessary."""
        cache_data = self.cache_obj.get(self.cache_key, {})
        self.cache_data.update(cache_data)
        try:
            self.process_thumbnails()
        except FileNotFoundError:
            logger.error("Image %s is missing.", self.src_path)
            timestamp = int(time.time())
            self.cache_data["next_slow_check"] = timestamp + self.src_slow_check_expiration
            self.cache_data["next_fast_check"] = timestamp + self.src_fast_check_expiration
            self.cache_changed = True
        except Exception as e:
            logger.error("Error while processing image %s [%s]", self.src_path, e)
            timestamp = int(time.time())
            self.cache_data["next_slow_check"] = timestamp + self.src_slow_check_expiration
            self.cache_data["next_fast_check"] = timestamp + self.src_fast_check_expiration
            self.cache_changed = True
        if self.cache_changed:
            self.cache_obj.set(self.cache_key, self.cache_data, self.src_cache_expiration)

    def process_thumbnails(self):
        """Update image metadata if required and create missing thumbnails."""
        old_cache_data = self.cache_data.copy()
        missing_data = any(self.cache_data[x] is None for x in self.default_cache_data)
        timestamp = int(time.time())
        if (
            missing_data
            or self.cache_data.get("next_slow_check") < timestamp
            or (self.cache_data.get("next_fast_check") < timestamp and self.get_mtime() > self.cache_data["mtime"])
        ):
            # we must recompute the cached data since either the data is missing
            # or the last slow check is too old
            # or the new fast check shows that the file has been modified
            self.update_image_data()
            self.cache_data["next_slow_check"] = timestamp + self.src_slow_check_expiration
            self.cache_data["next_fast_check"] = timestamp + self.src_fast_check_expiration
            self.cache_data["mtime"] = self.get_mtime()
            self.cache_changed = True
        must_recreate = any(
            self.cache_data[x] != old_cache_data[x] for x in ("width", "height", "filesize", "sha256", "mtime")
        )
        widths_to_create: Set[int] = set()
        for width in self.get_required_widths():
            if must_recreate or self.get_cache_key(width) not in self.cache_data:
                widths_to_create.add(width)
                self.cache_changed = True
        base_width = self.get_base_width()
        base_cache_key = self.get_cache_key(base_width)
        if must_recreate or base_cache_key not in self.cache_data:
            widths_to_create.add(base_width)
            self.cache_changed = True
        if widths_to_create:
            self.create_thumbnails(widths_to_create, force=must_recreate)
        for width in self.get_required_widths():
            self.paths_srcset[f"{width}w"] = self.cache_data[self.get_cache_key(width)]
        self.path_src = self.cache_data[base_cache_key]

    def get_mtime(self) -> int:
        """Return the modification time of the source image, when available."""
        try:
            dt: datetime.datetime = self.src_storage_obj.get_modified_time(self.src_path)
            return int(dt.timestamp())
        except NotImplementedError:
            return 0

    def get_base_width(self) -> int:
        """Return the base width of the displayed image."""
        if self.target_width is not None:
            return self.target_width
        elif self.target_height is not None:
            return math.ceil(self.target_height * self.cache_data["width"] / self.cache_data["height"])
        return self.cache_data["width"]

    def get_required_widths(self) -> Iterable[int]:
        """Return the list of required widths for the image."""
        if self.widths is None:
            self.widths = self.default_widths
        return [x for x in self.widths if x <= self.cache_data["width"]]

    def update_image_data(self):
        """Update the cache data for the image, allowing to check if thumbnails must be computed."""
        with self.src_storage_obj.open(self.src_path, "rb") as fd:
            img = Image.open(fd)
            self.cache_data["width"], self.cache_data["height"] = img.size
            sha256 = hashlib.sha256()
            fd.seek(0)
            for data in iter(lambda: fd.read(4096), b""):
                sha256.update(data)
            self.cache_data["filesize"] = fd.tell()
            self.cache_data["sha256"] = hashlib.sha256().hexdigest()
            img.close()

    def create_thumbnails(self, widths: Iterable[int], force: bool = False, write: bool = True):
        """Create the thumbnails for the all given widths."""
        if write:
            with self.src_storage_obj.open(self.src_path, "rb") as fd:
                img = Image.open(fd)
                for width in widths:
                    self.create_thumbnail(img, width, force=force, write=write)
                img.close()
        else:
            for width in widths:
                self.create_thumbnail(None, width, force=force, write=write)

    def get_target_path(self, width: int, height: int) -> str:
        """Return the target path for the given width."""
        return f"{self.cache_prefix}/{width}x{height}_{self.aspect_policy}/{self.src_path}.{self.fmt}"

    @classmethod
    def from_target_path(cls, dst_path: str):
        """Build a CachedImage from the URL path, if a thumbnail is missing."""
        if not (matcher := cls.target_path_re.match(dst_path)):
            raise Http404("Invalid path")
        cache_prefix = matcher.group("cache_prefix")
        for source in cls.sources.values():
            if cache_prefix == source["prefix"]:
                break
        else:
            raise Http404("Unkown image path.")
        src_storage = source["src_storage"]
        dst_storage = source["dst_storage"]
        cache = source["cache"]
        if not source.get("reversible"):
            raise Http404("Invalid image path.")
        width = int(matcher.group("width"))
        height = int(matcher.group("height"))
        if not (0 < height < 10000 and 0 < width < 10000):
            raise Http404("Invalid image size.")
        try:
            aspect_policy = AspectPolicy.get_policy(int(matcher.group("aspect_policy")))
        except ValueError:
            raise Http404("Invalid aspect policy.")
        src_path = matcher.group("src_path")
        src_path = os.path.normpath(src_path)
        if src_path.startswith("../") or src_path in {".", ".."}:
            raise Http404("Invalid source path.")
        fmt = matcher.group("fmt")
        if fmt not in cls.formats:
            raise Http404("Invalid format.")
        return cls(
            src_storage,
            dst_storage,
            cache,
            src_path,
            cache_prefix,
            height=height,
            width=width,
            widths=[],
            aspect_policy=aspect_policy,
            fmt=fmt,
        )

    @classmethod
    def get_cache_key(cls, width: int) -> str:
        """Return the cache key for the given width."""
        return f"{width}w"

    def create_thumbnail(
        self,
        img_src: Optional[Image],
        required_width: int,
        force: bool = False,
        write: bool = True,
    ):
        """Create a thumbnail for the given width and stores the new name in the cache."""
        src_width = self.cache_data["width"]
        src_height = self.cache_data["height"]
        width_cache_key = self.get_cache_key(required_width)
        aspect_ratio_height = math.ceil(required_width * src_height / src_width)
        if self.target_width and self.target_height:
            required_height = math.ceil(required_width * self.target_height / self.target_width)
        else:
            required_height = aspect_ratio_height
        dst_path = self.get_target_path(required_width, required_height)
        if width_cache_key not in self.cache_data:
            self.cache_data[width_cache_key] = dst_path
            self.cache_changed = True
        if not write:
            return
        exists = self.dst_storage_obj.exists(dst_path)
        if not force and exists:
            return
        logger.info("Creating thumbnail %s", dst_path)
        if aspect_ratio_height == required_height:
            # perfect fit: the target aspect ratio is the same as the source aspect ratio
            # always true when only one dimension is specified
            img = img_src.resize((required_width, required_height))
        elif aspect_ratio_height < required_height:
            # the image is not high enough
            if (
                self.aspect_policy == AspectPolicy.FORCE_HEIGHT_CROP
                or self.aspect_policy == AspectPolicy.FORCE_MIN_CROP
            ):
                tmp_width = math.ceil(required_height * src_width / src_height)
                img = img_src.resize((tmp_width, required_height))  # larger than expected
                half = (tmp_width - required_width) // 2
                img = img.crop((half, 0, required_width + half, required_height))  # we crop
            elif (
                self.aspect_policy == AspectPolicy.FORCE_WIDTH_CROP or self.aspect_policy == AspectPolicy.FORCE_MAX_FIT
            ):
                img = img_src.resize((required_width, aspect_ratio_height))  # shorter than expected
            else:  # if self.aspect_policy == AspectPolicy.FORCE_MAX_TRANSPARENT:
                r_img: Image = img_src.resize((required_width, aspect_ratio_height))  # shorter than expected
                img = Image.new("RGBA", (required_width, required_height), (255, 0, 0, 0))
                img.paste(r_img, (0, (required_height - aspect_ratio_height) // 2))
                r_img.close()
        else:
            # the image is not wide enough
            aspect_ratio_width = math.ceil(required_height * src_width / src_height)
            if self.aspect_policy == AspectPolicy.FORCE_HEIGHT_CROP or self.aspect_policy == AspectPolicy.FORCE_MAX_FIT:
                img = img_src.resize((aspect_ratio_width, required_height))  # thinner than expected
            elif (
                self.aspect_policy == AspectPolicy.FORCE_WIDTH_CROP or self.aspect_policy == AspectPolicy.FORCE_MIN_CROP
            ):
                tmp_height = math.ceil(required_width * src_height / src_width)
                img = img_src.resize((required_width, tmp_height))  # taller than expected
                half = (tmp_height - required_height) // 2
                img = img.crop((0, half, required_width, required_height + half))  # we crop
            else:  # if self.aspect_policy == AspectPolicy.FORCE_MAX_TRANSPARENT:
                r_img: Image = img_src.resize((aspect_ratio_width, required_height))
                img = Image.new("RGBA", (required_width, required_height), (255, 0, 0, 0))
                img.paste(r_img, ((required_width - aspect_ratio_width) // 2, 0))
                r_img.close()
        self.prepare_storage(dst_path)
        if exists:
            self.dst_storage_obj.delete(dst_path)
        with self.dst_storage_obj.open(dst_path, "wb") as fd:
            img.save(fd, format=self.fmt.upper())
        self.created_sizes.add(required_width)

    def prepare_storage(self, path):
        """Prepare the storage for the given path, creating directories if required."""
        if isinstance(self.dst_storage_obj, FileSystemStorage):
            full_path = self.dst_storage_obj.path(path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)

    @property
    def attr_src(self) -> str:
        """Return the src attribute for the <img> tag."""
        return self.url(self.path_src)

    @property
    def attr_srcset(self) -> str:
        """Return the srcset attribute for the <img> tag."""
        return ",".join(f"{self.url(path)} {size}" for size, path in self.paths_srcset.items())

    def as_html_tag(
        self,
        alt: str = "",
        crossorigin: str = "anonymous",
        loading: str = "lazy",
        referrerpolicy: str = "same-origin",
        **kwargs,
    ):
        """Return a URL tag with all required attributes."""
        if not self.path_src:
            return ""
        attrs = {"src": escape(self.attr_src)}
        if self.paths_srcset:
            attrs["srcset"] = escape(self.attr_srcset)
        attrs["alt"] = escape(alt)
        attrs["crossorigin"] = escape(crossorigin)
        attrs["loading"] = escape(loading)
        attrs["referrerpolicy"] = escape(referrerpolicy)
        if self.target_width:
            attrs["width"] = str(self.target_width)
        if self.target_height:
            attrs["height"] = str(self.target_height)
        for k, v in kwargs.items():
            k = k.replace("_", "-")
            if k == "class-":
                k = "class"
            attrs[escape(k)] = escape(v)
        attrs = " ".join(f'{k}="{v}"' for k, v in attrs.items())
        return mark_safe(f"<img {attrs}/>")  # noqa S308

    def log_error(self, request: Optional[HttpRequest]):
        """Log errors."""
        if request is None:
            logger.error("Image %s not found", self.src_path)
        else:
            request_path = request.path
            referer = request.META.get("HTTP_REFERER", "")
            logger.error("Image %s not found (referer: %s, URI: %s)", self.src_path, referer, request_path)



"""Custom form fields for the df_site app."""

import html
from functools import lru_cache
from typing import Dict, Iterable, List, Set

import nh3
from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.html import strip_tags
from django_ckeditor_5.widgets import CKEditor5Widget
from lxml_html_clean import autolink_html

# list of CKEditor5 plugins and associated HTML tags
BLOCK = "$block"
TAGS_BY_PLUGIN: Dict[str, Dict[str, List[str]]] = {
    "Alignment": {BLOCK: ["class"]},
    "BlockQuote": {"blockquote": []},
    "Bold": {"strong": []},
    "Code": {"code": []},
    "CodeBlock": {"pre": ["class"], "code": ["class"]},
    "Essentials": {"br": []},
    "Font": {"span": ["class"]},
    "Heading": {"h1": [], "h2": [], "h3": [], "h4": []},
    "Highlight": {"mark": ["class"]},
    "HorizontalLine": {"hr": []},
    "HtmlEmbed": {"div": ["class"]},
    "Image": {"figure": ["class"], "img": []},
    "ImageCaption": {"figcaption": []},
    "Indent": {BLOCK: ["class"]},
    "Italic": {"i": []},
    "Link": {"a": ["rel", "href"]},
    "List": {"ol": [], "ul": [], "li": []},
    "ListProperties": {"ol": [], "ul": [], "li": []},
    "MediaEmbed": {
        "figure": ["class"],
        "oembed": ["url"],
        "div": ["data-oembed-url"],
        "iframe": ["frameborder", "src", "allow"],
    },
    "Mention": {"span": ["class", "data-mention"]},
    "Paragraph": {"p": []},
    "Strikethrough": {"s": []},
    "Subscript": {"sub": []},
    "Superscript": {"sup": []},
    "Table": {
        "figure": ["class"],
        "table": [],
        "thead": [],
        "tbody": [],
        "tr": [],
        "td": ["colspan", "rowspan"],
        "th": ["colspan", "rowspan"],
    },
    "TableCaption": {"figcaption": ["data-placeholder"]},
    "TableCellProperties": {"td": ["class"]},
    "TableProperties": {"figure": ["class"], "table": ["class"]},
    "TodoList": {"ul": ["class"], "label": ["class"], "span": ["class"], "li": ["class"], "input": ["class"]},
    "Underline": {"u": []},
}
BLOCKS = {"blockquote", "pre", "div", "figure", "ol", "ul", "p", "table", "tr", "td", "th"}


@lru_cache
def get_allowed_tags(config_name: str) -> Set[str]:
    """Return the list of allowed HTML tags for a given CKEditor 5 configuration."""
    return set(get_allowed_attributes(config_name))


def get_allowed_attributes(config_name: str) -> Dict[str, Set[str]]:
    """Return the list of allowed HTML tags and attributes for a given CKEditor 5 configuration."""
    config = settings.CKEDITOR_5_CONFIGS.get(config_name, {})
    plugins: Iterable[str] = config.get("plugins", TAGS_BY_PLUGIN.keys())
    allowed_attributes: Dict[str, Set[str]] = {}
    for plugin in plugins:
        for key, attributes in TAGS_BY_PLUGIN.get(plugin, {}).items():
            if key == BLOCK:
                keys = BLOCKS
            else:
                keys = [key]
            for key_ in keys:
                allowed_attributes.setdefault(key_, set())
                allowed_attributes[key_] |= set(attributes)
    return allowed_attributes


class CKEditor5Field(forms.CharField):
    """A form field using CKEditor 5 as text editor."""

    default_ckeditor5_config = "default"

    def __init__(self, *args, config_name=None, **kwargs) -> None:
        """Initialize the field."""
        self.ckeditor5_config = config_name or self.default_ckeditor5_config
        self.required_ = kwargs.get("required", True)
        kwargs["required"] = False
        css_class = "django_ckeditor_5"
        self.is_inline_widget = "inline" in self.ckeditor5_config
        if self.is_inline_widget:
            css_class += " ck-editor__singleline"
        kwargs["widget"] = CKEditor5Widget(attrs={"class": css_class}, config_name=self.ckeditor5_config)
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        """Convert the value to a string."""
        allowed_tags = get_allowed_tags(self.ckeditor5_config)
        allowed_attributes = get_allowed_attributes(self.ckeditor5_config)
        if value is None:
            raise ValidationError(self.error_messages["required"], code="required")
        elif len(value) <= 50 and self.required_:
            striped = html.unescape(strip_tags(value)).strip()
            if not striped:
                raise ValidationError(self.error_messages["required"], code="required")
        if self.is_inline_widget and "p" in allowed_tags:
            allowed_tags.remove("p")
        value = nh3.clean(value, strip_comments=True, tags=allowed_tags, attributes=allowed_attributes)
        value = autolink_html(value)
        value = value.replace("\n", " ")
        return super().to_python(value)

    def prepare_value(self, value):
        """Prepare the value for the widget."""
        if value is None:
            return ""
        return value


class InlineCKEditor5Field(CKEditor5Field):
    """A form field using CKEditor 5 as inline text editor."""

    default_ckeditor5_config = "inline"


class InlineLinkCKEditor5Field(CKEditor5Field):
    """A form field using CKEditor 5 as inline text editor."""

    default_ckeditor5_config = "inline_link"



"""Allow to use "python3 -m df_site" to run the manage command."""

import multiprocessing
import os
import sys
import warnings

from df_config.manage import DEFAULT_SETTINGS_MODULE, manage, set_env


def setup(module_name: str = None, settings_module=DEFAULT_SETTINGS_MODULE):
    """Setup the Django environment."""
    set_env(module_name=module_name, settings_module=settings_module)
    import django

    django.setup()
    from django.conf import settings

    return settings


def main(module_name: str = "df_site"):
    """Run the manage command."""
    if not sys.warnoptions:
        warnings.simplefilter("default")  # Change the filter in this process
        os.environ["PYTHONWARNINGS"] = "default"  # Also affect subprocesses
    # avoid "django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet." during parallel testing
    multiprocessing.set_start_method("fork")

    # required for parallel testing on macOS and Python 3.8+
    os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"
    manage(module_name=module_name)


if __name__ == "__main__":
    """Allow to use "python3 -m df_site"."""
    main()



"""Manage user settings."""

from functools import lru_cache
from typing import Any, Optional

from cookie_consent.util import get_cookie_value_from_request
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models import Field
from django.forms import Widget
from django.http import HttpRequest, HttpResponse

from df_site.models import AbstractPreferences


@lru_cache
def get_user_instance() -> AbstractPreferences:
    """Return a blank user instance, required for validation purposes."""
    return get_user_model()()


def get_user_setting(
    attr: str, user: Optional[AbstractPreferences] = None, request: Optional[HttpRequest] = None
) -> Any:
    """Return the specified user setting.

    If the user is given, the value is read from the user instance.
    If the cookie provides a value, it is validated and used.
    Otherwise, the default value is returned.
    """
    if user:
        return getattr(user, attr)
    elif request and request.user.is_authenticated:
        return getattr(request.user, attr)
    user_instance = get_user_instance()
    opts = user_instance._meta
    field = opts.get_field(attr)
    raw_value = request.COOKIES.get(attr) if request else None
    value = field.default
    if raw_value is not None:
        try:
            value = field.formfield().clean(raw_value)
        except ValidationError:
            pass
    return value


def set_user_setting(
    attr: str,
    value: Any,
    user: Optional[AbstractPreferences] = None,
    request: Optional[HttpRequest] = None,
    response: Optional[HttpResponse] = None,
) -> None:
    """Set the user's preferences."""
    if user:
        setattr(user, attr, value)
        user.save()
    elif request and request.user.is_authenticated:
        setattr(request.user, attr, value)
        request.user.save()
    elif get_cookie_value_from_request(request, "user-settings"):
        user_instance = get_user_instance()
        opts = user_instance._meta
        field: Field = opts.get_field(attr)
        widget: Widget = field.formfield().hidden_widget()
        raw_value: str = widget.format_value(value)
        response.set_cookie(attr, raw_value, samesite="Strict", secure=settings.USE_SSL)



"""Custom template tags for the df_site app."""

import datetime
import re
from html import escape
from typing import Dict, List, Optional, Union

from django import template
from django.conf import settings
from django.contrib.admin.views.main import PAGE_VAR
from django.core.paginator import Paginator
from django.template.base import kwarg_re
from django.template.defaultfilters import date
from django.template.defaulttags import URLNode
from django.template.exceptions import TemplateSyntaxError
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from df_site.components.base import Component
from df_site.constants import BRAND_ICONS, INT_RE

register = template.Library()


@register.filter()
def abs_url(value):
    """Given a relative URL, return an absolute URL."""
    if value and value.startswith("/"):
        base_url = settings.SERVER_BASE_URL
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        return mark_safe(f"{base_url}{value}")  # noqa
    return value


class AbsoluteURLNode(URLNode):
    """A template node that renders an absolute URL."""

    def render(self, context):
        """Render the URL node prefixed with the server base URL."""
        rendered = super().render(context)
        return abs_url(rendered)


@register.simple_tag(takes_context=True)
def paginate_qs(context, queryset, per_page=10, page_attr="page"):
    """Paginate a queryset and store the pagination in the context."""
    paginator = Paginator(queryset, per_page)
    page_number = context["request"].GET.get(page_attr, "1")
    if INT_RE.match(page_number):
        page = paginator.get_page(int(page_number))
    else:
        page = paginator.get_page(1)
    return page


@register.tag(name="abs_url")
def abs_url_tag(parser, token):
    """Like the default Django's url tag, but return absolute URLs."""
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError(f"'{bits[0]}' takes at least one argument, a URL pattern name.")
    viewname = parser.compile_filter(bits[1])
    args = []
    kwargs = {}
    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == "as":
        asvar = bits[-1]
        bits = bits[:-2]

    for bit in bits:
        match = kwarg_re.match(bit)
        if not match:
            raise TemplateSyntaxError("Malformed arguments to url tag")
        name, value = match.groups()
        if name:
            kwargs[name] = parser.compile_filter(value)
        else:
            args.append(parser.compile_filter(value))
    return AbsoluteURLNode(viewname, args, kwargs, asvar)


@register.filter
def fa6_allauth_icon(provider_id):
    """Return the font-awesome icon name for a given allauth provider."""
    return BRAND_ICONS.get(provider_id, "key")


@register.simple_tag
def fa6_icon(
    name,
    prefix="fa",
    large: Union[int, bool] = False,
    fixed: bool = False,
    spin: bool = False,
    li: bool = False,
    rotate: Union[int, bool] = None,
    border: bool = False,
    color: str = None,
):
    """Add font-awesome icons in your HTML code."""
    if isinstance(large, int) and 2 <= large <= 5:
        large = f" fa-{large:d}x"
    elif large:
        large = " fa-lg"
    else:
        large = ""
    content = '<i class="{prefix} fa-{name}{large}{fixed}{spin}{li}{rotate}{border}{color}"></i>'.format(
        prefix=prefix,
        name=name,
        large=large,
        fixed=" fa-fw" if fixed else "",
        spin=" fa-spin" if spin else "",
        li=" fa-li" if li else "",
        rotate=f" fa-rotate-{rotate}" if rotate else "",
        border=" fa-border" if border else "",
        color=f"text-{color}" if color else "",
    )
    return mark_safe(content)  # noqa


@register.simple_tag
def bootstrap_breadcrumb(
    breadcrumbs: List[Dict[str, Union[str, bool]]], last_active: bool = True, extra_classes: str = ""
):
    """Render an breadcrumb.

    **Tag name**::

        bootstrap_breadcrumb

    **Parameters**::

        breadcrumbs
            List of dictionaries with the following keys:
                `link`: URL to link to
                `title`: Text to display
                `active`: boolean, element is active, meaning that the link is not displayed

        last_active
            boolean, last element is the active one

            :default: ``True``

        extra_classes
            string, extra CSS classes for elements

            :default: ""

    **Usage**::

        {% bootstrap_breadcrumb [{content}] %}

    **Example**::

        {% bootstrap_breadcrumb crumbs %}
    """
    context = {
        "breadcrumbs": breadcrumbs,
        "last_active": last_active,
        "extra_classes": extra_classes,
    }
    content = render_to_string("django_bootstrap5/breadcrumb.html", context)
    return mark_safe(content)  # noqa S308


@register.simple_tag(takes_context=True)
def component(context, comp: Component, **kwargs):
    """Render a HTML component in a template."""
    text = comp.render(context, **kwargs)
    return mark_safe(text)  # noqa S308


@register.simple_tag
def component_list_url(cl, page: Optional[int] = None):
    """Generate an individual page index link in a paginated list."""
    kwargs = {}
    if page is not None:
        kwargs[PAGE_VAR] = page
    return cl.get_query_string(kwargs)


auto_pill_re = re.compile(r"^(.*)\((\d+)\)\s*$")


@register.simple_tag
def component_auto_badge(
    value: str, color: str = "primary", extra_class: str = "rounded-pill", regex: re.Pattern = auto_pill_re
):
    """Automatically extracts suffixes between parenthesis `(.*)` and put them into badges."""
    if hasattr(value, "__html__"):
        raw_value = str(value)
    else:
        raw_value = escape(str(value))
    if match := regex.match(raw_value):
        prefix = match.group(1)
        suffix = match.group(2)
        new_value = f'{prefix}<span class="badge text-bg-{color} {extra_class}">{suffix}</span>'
        value = mark_safe(new_value)  # noqa S308
    return value


@register.simple_tag
@register.filter
def schema_date(
    value_: Optional[Union[datetime.date, datetime.datetime, datetime.timedelta]], fmt_: Optional[str] = None, **kwargs
):
    """Show a date in a <time></time> element.

    Can be used as a filter or as a tag. In the latter case, value and format are the first two arguments.
    Extra arguments are passed as attributes to the generated HTML.

    `{% schema_date value_ fmt_ itemprop="start" %}` is rendered as
        `<time itemprop="start" datetime="date">date</time>`
    `{{ value_|schema_date:fmt_ }}` is rendered as `<time datetime="date">date</time>`
    """
    content = date(value_, fmt_)
    if isinstance(value_, datetime.date):
        kwargs["datetime"] = value_.strftime("%Y-%m-%d")
    elif isinstance(value_, datetime.datetime):
        kwargs["datetime"] = value_.isoformat()
    elif isinstance(value_, datetime.timedelta):
        attr = "P"
        ts = value_.total_seconds()
        days = int(ts // 86400)
        ts %= 86400
        attr += f"{days}D"
        hours = int(ts // 3600)
        ts %= 3600
        attr += f"{hours}H"
        minutes = int(ts // 60)
        ts %= 60
        attr += f"{minutes}M{ts}S"
        kwargs["datetime"] = attr
    else:
        return content
    escaped = {escape(k): escape(v) for k, v in kwargs.items()}
    attr = " ".join(f'{k}="{v}"' for k, v in escaped.items())
    msg = f"<time {attr}>{content}</time>"
    return mark_safe(msg)  # noqa S308



"""Forms for the users app."""

from django import forms
from django.contrib.auth import get_user_model
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from django_recaptcha.fields import ReCaptchaField
from django_recaptcha.widgets import ReCaptchaV2Checkbox


class UserSettingsForm(forms.ModelForm):
    """Form for the user settings page."""

    class Meta:
        """Meta options for the form."""

        model = get_user_model()
        fields = [
            "first_name",
            "last_name",
            "color_theme",
            "email_notifications",
            "display_online",
        ]


class ReCaptchaForm(forms.Form):
    """Form for the reCAPTCHA field."""

    captcha = ReCaptchaField(widget=ReCaptchaV2Checkbox, label=_("I'm not a robot"))

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def signup(self, request: HttpRequest, user):
        """Configure the user as required by django-allauth."""
        return user



"""Simulate a classical Django admin_site with similare URLs."""

import warnings
from collections import defaultdict
from typing import Optional, Type

from django.contrib.admin import ModelAdmin
from django.db import models


class ModelURLRegistry:
    """A registry for models and their URLs."""

    def __init__(self):
        """Initialize the registry."""
        self.all_models = defaultdict(dict)

    def register_model(self, app_label, model):
        """Register a model in the registry."""
        # Since this method is called when models are imported, it cannot
        # perform imports because of the risk of import loops. It mustn't
        # call get_app_config().
        # noinspection PyProtectedMember
        model_name = model._meta.model_name
        app_models = self.all_models[app_label]
        if model_name in app_models:
            existing = app_models[model_name]
            if model.__name__ == existing.__name__ and model.__module__ == existing.__module__:
                warnings.warn(
                    f"Model '{app_label}.{model_name}' was already registered. Reloading models is not "
                    "advised as it can lead to inconsistencies, most notably with "
                    "related models.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            else:
                msg = f"Conflicting '{model_name}' models in application '{app_label}': {existing} and {model}."
                raise RuntimeError(msg)
        app_models[model_name] = model

    def unregister_model(self, app_label, model_name):
        """Unregister a model from the registry."""
        if app_label in self.all_models:
            if model_name in self.all_models[app_label]:
                del self.all_models[app_label][model_name]
            if len(self.all_models[app_label]) == 0:
                del self.all_models[app_label]

    # noinspection PyMethodMayBeStatic
    def get_model_admin(self, model: Type[models.Model]) -> Optional[ModelAdmin]:
        """Return an empty model admin, for the sake of compatibility."""
        return None


default_registry = ModelURLRegistry()



"""Groups all views related to the users."""

from django.conf import settings
from django.contrib import messages
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme, urlencode
from django.utils.translation import gettext as _
from django.views.decorators.cache import never_cache
from django.views.generic import FormView

from df_site.user_settings import set_user_setting
from df_site.users.forms import UserSettingsForm


class UserSettingsView(FormView):
    """View for the user settings page."""

    template_name = "users/settings.html"
    form_class = UserSettingsForm

    def get_success_url(self):
        """Get the URL to redirect to after a successful form submission."""
        return reverse("users:settings")

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.request.user
        return kwargs

    def form_valid(self, form):
        """Save the form data and redirect to the success URL."""
        form.save()
        messages.success(self.request, _("Your settings have been saved."))
        return super().form_valid(form)


@never_cache
def theme_switch(request: HttpRequest) -> HttpResponse:
    """Change the color theme of the site."""
    # noinspection PyTypeChecker
    current_theme: str = request.GET.get("current", "auto")
    # noinspection PyTypeChecker
    redirect_to = sanitize_redirection(request)

    if settings.DF_SITE_THEMES:
        next_theme, __, next_icon = settings.DF_SITE_THEMES[-1]
        for theme, label, icon in settings.DF_SITE_THEMES:
            if theme == current_theme:
                break
            next_theme, next_icon = theme, icon
    else:
        next_theme, next_icon = "auto", "toggle-on"
    next_url = reverse("users:theme_switch")
    args = urlencode({"next": redirect_to, "current": next_theme})
    result = {"theme": next_theme, "icon": next_icon, "href": f"{next_url}?{args}"}
    if request.META["HTTP_ACCEPT"] == "application/json":
        response = JsonResponse(result)
    else:
        response = HttpResponseRedirect(redirect_to)
    set_user_setting("color_theme", next_theme, request=request, response=response)
    return response


def sanitize_redirection(request, param="next"):
    """Sanitize the redirection URL, only keeping allowed hosts."""
    redirect_to: str = request.GET.get(param, "/")
    if not url_has_allowed_host_and_scheme(
        url=redirect_to,
        allowed_hosts={request.get_host()},
        require_https=request.is_secure(),
    ):
        redirect_to = "/"
    return redirect_to



"""Add an image to the media storage."""

import pathlib

from django.core.files.storage import Storage, storages
from django.core.management import BaseCommand
from django.urls import reverse

from df_site.templatetags.images import AspectPolicy, CachedImage


class Command(BaseCommand):
    """Add an image to the media storage."""

    help = "Add an image to the media storage."

    def add_arguments(self, parser):
        """Add arguments to the command."""
        parser.add_argument("path", type=pathlib.Path)
        parser.add_argument("--storage", choices=["default", "staticfiles"], default="default")
        parser.add_argument("--name", type=str, default=None)
        parser.add_argument("--height", type=int, default=None)
        parser.add_argument("--width", type=int, default=None)
        parser.add_argument("--format", type=str, default="webp", choices=CachedImage.formats)
        parser.add_argument("--overwrite", action="store_true", default=False)
        parser.add_argument("--print", action="store_true", default=False)
        parser.add_argument(
            "--policy",
            type=int,
            default=AspectPolicy.FORCE_MAX_TRANSPARENT.value,
            choices=[x.value for x in AspectPolicy],
        )

    def handle(self, *args, **options):
        """Add an image to the media storage and generate thumbnails."""
        aspect_policy = AspectPolicy.get_policy(options["policy"])
        src_path: pathlib.Path = options["path"]
        storage: str = options["storage"]
        name: str = options["name"] or src_path.name
        # we need to check if the file exists
        if not src_path.exists():
            raise FileNotFoundError(f"File {src_path} does not exist.")
        # we need to check if the file exists in the destination storage
        storage_obj: Storage = storages[storage]
        exists = storage_obj.exists(name)
        cache_engine = "locmem"
        fmt = options["format"]
        if exists and not options["overwrite"]:
            self.stderr.write(f"File {name} already exists in storage {storage}.")
            return
        elif exists:
            storage_obj.delete(name)
        # we save the file to the destination storage
        with src_path.open("rb") as src_file:
            storage_obj.save(name, src_file)
        #
        img = CachedImage.from_target_path(f"T/600x600_4/{name}.{fmt}")
        img.process()
        self.stdout.write(f"Generated thumbnails: {img.created_sizes}.")
        if options["print"]:
            self.stdout.write(img.as_html_tag())

        def url(p):
            """Return the URL of the thumbnail."""
            return reverse("thumbnails", kwargs={"path": p})

        img = CachedImage(
            CachedImage.sources["media"]["src_storage"],
            CachedImage.sources["media"]["dst_storage"],
            cache_engine,
            name,
            CachedImage.sources["media"]["prefix"],
            url=url,
            height=options["height"],
            width=options["width"],
            widths=None,
            aspect_policy=aspect_policy,
            fmt=fmt,
            write_thumbnails=True,
        )
        img.process()
        self.stdout.write(f"File {name} saved to storage {storage}.")
        self.stdout.write(f"Generated thumbnails: {img.created_sizes}.")
        if options["print"]:
            self.stdout.write(img.as_html_tag())
        img = CachedImage(
            CachedImage.sources["media"]["src_storage"],
            CachedImage.sources["media"]["dst_storage"],
            cache_engine,
            name,
            CachedImage.sources["media"]["prefix"],
            url=url,
            height=options["height"],
            width=options["width"],
            widths=[320, 480, 640, 1024, 2048],
            aspect_policy=aspect_policy,
            fmt=fmt,
            write_thumbnails=True,
        )
        img.process()
        self.stdout.write(f"File {name} saved to storage {storage}.")
        self.stdout.write(f"Generated thumbnails: {img.created_sizes}.")
        if options["print"]:
            self.stdout.write(img.as_html_tag())



"""Models for the df_site app."""

from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from df_site.model_fields import CKEditor5CharField, CKEditor5TextField


class AlertRibbon(models.Model):
    """Model for alert messages that are shown to users."""

    LEVELS = (
        ("info", _("Info")),
        ("success", _("Success")),
        ("warning", _("Warning")),
        ("danger", _("Danger")),
        ("primary", _("Primary")),
        ("secondary", _("Secondary")),
        ("dark", _("Dark")),
        ("light", _("Light")),
        ("tricolor", _("Tricolor")),
    )
    TOP_LEFT = 0
    BOTTOM_LEFT = 1
    TOP_RIGHT = 2
    BOTTOM_RIGHT = 3
    TOP_CENTER = 4
    BOTTOM_CENTER = 5
    POSITIONS = (
        (TOP_LEFT, _("Top left")),
        (BOTTOM_LEFT, _("Bottom left")),
        (TOP_RIGHT, _("Top right")),
        (BOTTOM_RIGHT, _("Bottom right")),
        (TOP_CENTER, _("Top center")),
        (BOTTOM_CENTER, _("Bottom center")),
    )

    color = models.CharField(max_length=10, choices=LEVELS, default="info", db_index=True, verbose_name=_("Color"))
    message = CKEditor5TextField(verbose_name=_("Message"), blank=True, default="")
    url = models.URLField(null=True, blank=True, verbose_name=_("URL"))  # noqa: DJ001
    summary = CKEditor5CharField(max_length=100, verbose_name=_("Summary"), db_index=True)
    start_date = models.DateTimeField(null=True, blank=True, db_index=True, verbose_name=_("Start date"))
    end_date = models.DateTimeField(null=True, blank=True, db_index=True, verbose_name=_("End date"))
    is_active = models.BooleanField(default=True, db_index=True, verbose_name=_("Is active"))
    position = models.PositiveSmallIntegerField(
        default=TOP_RIGHT, choices=POSITIONS, db_index=True, verbose_name=_("Position")
    )

    class Meta:
        """Meta options for the model."""

        ordering = ("-start_date",)
        verbose_name = _("Alert ribbon")
        verbose_name_plural = _("Alert ribbons")

    def __str__(self):
        """Return the string representation of the alert message."""
        return self.summary

    def get_absolute_url(self):
        """Return the URL for the alert message."""
        return reverse("ribbon", kwargs={"pk": self.pk})

    @property
    def html_tag(self):
        """Return the HTML tag for the alert message."""
        if self.position in (self.TOP_CENTER, self.BOTTOM_CENTER):
            return "div"
        elif self.url:
            return "a"
        return "span"

    @property
    def css_classes(self):
        """Return the CSS classes for the alert message."""
        if self.position == self.TOP_CENTER:
            classes = f"container-fluid btn btn-{self.color} "
        elif self.position == self.BOTTOM_CENTER:
            classes = f"container-fluid btn btn-{self.color} fixed-bottom "
        else:
            classes = f"alert-ribbon fixed alert-ribbon-{self.color} "
        if self.position in (self.TOP_LEFT, self.BOTTOM_LEFT):
            classes += "left-"
        elif self.position in (self.TOP_RIGHT, self.BOTTOM_RIGHT):
            classes += "right-"
        elif self.position in (self.TOP_CENTER, self.BOTTOM_CENTER):
            classes += "center-"
        if self.position in (self.TOP_LEFT, self.TOP_RIGHT, self.TOP_CENTER):
            classes += "top"
        elif self.position in (self.BOTTOM_LEFT, self.BOTTOM_RIGHT, self.BOTTOM_CENTER):
            classes += "bottom"
        return classes


class AbstractPreferences(models.Model):
    """User preferences for the df_site app."""

    COLOR_THEMES = {x[0]: x[1] for x in settings.DF_SITE_THEMES}
    color_theme = models.CharField(
        max_length=10,
        default=settings.DF_SITE_THEMES[0][0],
        db_index=True,
        choices=COLOR_THEMES,
        verbose_name=_("Color theme"),
    )
    display_online = models.BooleanField(default=False, verbose_name=_("Display online status"), db_index=True)
    email_notifications = models.BooleanField(
        default=False, verbose_name=_("Receive notifications by email"), db_index=True
    )

    class Meta:
        """Meta options for the model."""

        abstract = True


class PreferencesUser(AbstractUser, AbstractPreferences):
    """User model for the df_site app."""

    class Meta(AbstractUser.Meta):
        """Meta options for the model."""

        swappable = "AUTH_USER_MODEL"

    def __str__(self):
        """Return the string representation of the user."""
        return super().__str__()



"""Custom template tags for generating Subresource Integrity hashes for Pipeline assets."""

import base64
import hashlib
from functools import lru_cache

from django import template
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from pipeline.templatetags.pipeline import JavascriptNode, StylesheetNode
from pipeline.utils import guess_type

register = template.Library()


def get_sri(path, method=None):
    """Generate a Subresource Integrity hash for the given file."""
    if method in {"sha256", "sha384", "sha512"} and staticfiles_storage.exists(path):
        with staticfiles_storage.open(path) as fd:
            h = getattr(hashlib, method)()
            for data in iter(lambda: fd.read(16384), b""):
                h.update(data)
        hashed = base64.b64encode(h.digest()).decode()
        return f"{method}-{hashed}"
    return None


if not settings.DEBUG:
    get_sri = lru_cache(maxsize=1024)(get_sri)


class SRIJavascriptNode(JavascriptNode):
    """Render a <script> tag with a SRI hash for the given group."""

    def render_js(self, package, path):
        """Render the JS tag with SRI hash."""
        template_name = package.template_name or "pipeline/js.html"
        context = package.extra_context
        url = mark_safe(staticfiles_storage.url(path))  # noqa
        context.update(
            {
                "type": guess_type(path, "text/javascript"),
                "url": url,
                "crossorigin": "anonymous",
                "integrity": get_sri(path, method=package.config.get("integrity")),
            }
        )
        return render_to_string(template_name, context)


# noinspection PyUnusedLocal
@register.tag
def sri_javascript(parser, token):
    """Generate a <script> tag with a SRI hash for the given group."""
    try:
        tag_name, name = token.split_contents()
    except ValueError:
        tag_name = token.split_contents()[0]
        raise template.TemplateSyntaxError(
            f"{tag_name!r} requires exactly one argument: the name of a group in the PIPELINE.JAVASCRIPT setting"
        )
    return SRIJavascriptNode(name)


class SRIStylesheetNode(StylesheetNode):
    """Render a <link> tag with a SRI hash for the given group."""

    def render_css(self, package, path):
        """Render the CSS tag with SRI hash."""
        template_name = package.template_name or "pipeline/css.html"
        context = package.extra_context
        url = mark_safe(staticfiles_storage.url(path))  # noqa
        context.update(
            {
                "type": guess_type(path, "text/css"),
                "url": url,
                "crossorigin": "anonymous",
                "integrity": get_sri(path, method=package.config.get("integrity")),
            }
        )
        return render_to_string(template_name, context)


# noinspection PyUnusedLocal
@register.tag
def sri_stylesheet(parser, token):
    """Generate a <link> tag with a SRI hash for the given group."""
    try:
        tag_name, name = token.split_contents()
    except ValueError:
        tag_name = token.split_contents()[0]
        raise template.TemplateSyntaxError(
            f"{tag_name!r} requires exactly one argument: the name of a group in the PIPELINE.STYLESHEET setting"
        )
    return SRIStylesheetNode(name)



"""Set default topics for websockets."""

from df_websockets.tasks import set_websocket_topics
from django.http import HttpRequest, HttpResponse


def websocket_middleware(get_response):
    """Set default topics for websockets, when not set yet."""

    def middleware(request: HttpRequest):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        response: HttpResponse = get_response(request)
        # noinspection PyUnresolvedReferences
        if hasattr(request, "has_websocket_topics") and not request.has_websocket_topics:
            set_websocket_topics(request)
        return response

    return middleware



# Generated by Django 5.1 on 2024-08-23 16:32

import df_site.model_fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("df_site", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="alertribbon",
            name="message",
            field=df_site.model_fields.CKEditor5TextField(blank=True, default="", verbose_name="Message"),
        ),
        migrations.AlterField(
            model_name="alertribbon",
            name="summary",
            field=df_site.model_fields.CKEditor5CharField(db_index=True, max_length=100, verbose_name="Summary"),
        ),
    ]



"""Define custom model fields with the CKEditor5 widget."""

from django.db import models

from df_site import form_fields


class CKEditor5TextField(models.TextField):
    """Use a CKEditor5 widget for a TextField."""

    def formfield(self, **kwargs):
        """Replace the default widget with a CKEditor5 widget."""
        defaults = {"form_class": form_fields.CKEditor5Field}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class CKEditor5CharField(models.CharField):
    """Use a CKEditor5 widget for a CharField."""

    def formfield(self, **kwargs):
        """Replace the default widget with a CKEditor5 widget."""
        defaults = {"form_class": form_fields.InlineCKEditor5Field}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class CKEditor5LinkCharField(models.CharField):
    """Use a CKEditor5 widget for a CharField, allowing to write URLs."""

    def formfield(self, **kwargs):
        """Replace the default widget with a CKEditor5 widget."""
        defaults = {"form_class": form_fields.InlineLinkCKEditor5Field}
        defaults.update(kwargs)
        return super().formfield(**defaults)



"""Components for displaying a list of items in a table."""

from typing import Callable, List, Optional, Tuple, Type, Union

from django.contrib.admin import ListFilter, ShowFacets, SimpleListFilter, widgets
from django.contrib.admin.options import IS_FACETS_VAR, IS_POPUP_VAR
from django.contrib.admin.templatetags.admin_list import date_hierarchy as raw_date_hierarchy
from django.contrib.admin.templatetags.admin_list import result_headers, results
from django.contrib.admin.utils import get_fields_from_path, lookup_spawns_duplicates
from django.contrib.admin.views.main import ALL_VAR, ORDER_VAR, PAGE_VAR, SEARCH_VAR, ChangeList
from django.core.exceptions import FieldDoesNotExist
from django.core.paginator import Paginator
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.db.models.functions.text import Substr, Upper
from django.http import HttpRequest
from django.utils.http import urlencode
from django.utils.text import smart_split, unescape_string_literal
from django.utils.translation import gettext as _

from df_site.components.base import ModelComponent


class ModelListChangeList(ChangeList):
    """Customized ChangeList for ModelListComponent."""

    formset = None
    paginator: Paginator

    def __init__(self, request: HttpRequest, *args, **kwargs):
        """Initialize the ChangeList."""
        super().__init__(request, *args, **kwargs)
        self.request: HttpRequest = request

    def url_for_result(self, result: models.Model):
        """Return the URL for a result, trying to use the get_absolute_url method."""
        if hasattr(result, "get_absolute_url"):
            url = result.get_absolute_url()
        else:
            url = f"/{self.opts.app_label}/{self.opts.model_name}/{result.pk}/show/"
        return url


class ModelListComponent(ModelComponent):
    """A component that can display a list as table, with filters and a search bar."""

    page_var = PAGE_VAR
    all_var = ALL_VAR
    order_var = ORDER_VAR
    search_var = SEARCH_VAR
    changelist_filters_var = "_changelist_filters"
    list_editable: List[str] = []

    def __init__(
        self,
        model: Type[models.Model],
        base_template: str = "list.html",
        list_select_related: Optional[List[str]] = None,
        list_display: List[str] = None,
        list_display_links: List[str] = None,
        list_filter: List[Union[str, Tuple[str, Type[ListFilter]], Callable]] = None,
        date_hierarchy: Optional[str] = None,
        search_fields: List[str] = None,
        list_per_page: int = 20,
        list_max_show_all: int = 200,
        sortable_by: List[str] = None,
        search_help_text: Optional[str] = None,
        show_facets: ShowFacets = ShowFacets.ALLOW,
        show_full_result_count: bool = True,
        ordering: List[str] = None,
        pagination_on_top: bool = True,
        pagination_on_bottom: bool = True,
        filters_on_right: bool = True,
        filters_title: str = _("Filters"),
    ):
        """Create a new list component."""
        super().__init__(model, base_template)
        self.list_select_related: Optional[List[str]] = list_select_related
        self.list_display: List[str] = list_display or ["__str__"]
        self.list_display_links: List[str] = list_display_links or []
        self.list_filter: List[Union[str, Tuple[str, Type[ListFilter], Callable]]] = list_filter or []
        self.date_hierarchy: Optional[str] = date_hierarchy
        self.search_fields: List[str] = search_fields or []
        self.list_per_page: int = list_per_page
        self.list_max_show_all: int = list_max_show_all
        self.sortable_by: List[str] = sortable_by or []
        self.search_help_text: Optional[str] = search_help_text
        self.show_facets: ShowFacets = show_facets
        self.show_full_result_count: bool = show_full_result_count
        self.ordering: List[str] = ordering or []
        self.pagination_on_top: bool = pagination_on_top
        self.pagination_on_bottom: bool = pagination_on_bottom
        self.filters_on_right: bool = filters_on_right
        self.filters_title: str = filters_title

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_change_list_class(self, request: HttpRequest, **kwargs) -> Type[ChangeList]:
        """Return the ChangeList class to use for this component."""
        return ModelListChangeList

    # noinspection PyUnusedLocal
    def get_change_list(self, request: HttpRequest, **kwargs) -> ChangeList:
        """Return the ChangeList instance to use for this component."""
        cls = self.get_change_list_class(request, **kwargs)
        return cls(
            request,
            self.model,
            self.get_list_display(request, **kwargs),
            self.get_list_display_links(request, **kwargs),
            self.get_list_filter(request, **kwargs),
            self.get_date_hierarchy(request, **kwargs),
            self.get_search_fields(request, **kwargs),
            self.list_select_related,
            self.list_per_page,
            self.list_max_show_all,
            self.list_editable,
            ModelAdminWrapper(self, kwargs),
            self.sortable_by,
            self.get_search_help_text(request, **kwargs),
        )

    # noinspection PyUnusedLocal
    def get_date_hierarchy(self, request: HttpRequest, **kwargs):
        """Return the date hierarchy, depending on the request."""
        return self.date_hierarchy

    # noinspection PyUnusedLocal
    def get_search_help_text(self, request: HttpRequest, **kwargs):
        """Return the text displayed before the search input."""
        return self.search_help_text

    # noinspection PyUnusedLocal
    def get_list_display_links(self, request: HttpRequest, **kwargs):
        """Return the list of fields with the link to the object."""
        return self.list_display_links

    # noinspection PyUnusedLocal
    def get_list_display(self, request: HttpRequest, **kwargs):
        """Return the list of fields to display."""
        return self.list_display

    def get_queryset(self, request: HttpRequest, **kwargs) -> models.QuerySet:
        """Return the queryset to use for this component."""
        qs = self.get_base_queryset(request, **kwargs)
        if self.list_select_related is True:
            qs = qs.select_related()
        elif self.list_select_related:
            qs = qs.select_related(*self.list_select_related)
        return qs

    def get_preserved_filters(self, request: HttpRequest):
        """Return the preserved filters querystring."""
        # noinspection PyArgumentList
        preserved_filters = request.GET.urlencode()
        if preserved_filters:
            return urlencode({self.changelist_filters_var: preserved_filters})
        return ""

    # noinspection PyUnusedLocal
    def to_field_allowed(self, request, to_field):
        """Mimic the behavior of the ModelAdmin."""
        try:
            field = self.opts.get_field(to_field)
        except FieldDoesNotExist:
            return False

        return field.primary_key

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def has_change_permission(self, request: HttpRequest, obj=None):
        """Mimic the behavior of the ModelAdmin."""
        return False

    def lookup_allowed(self, lookup, value, request: HttpRequest = None, **kwargs):
        """Mimic the behavior of the ModelAdmin."""
        model = self.model
        # Check FKey lookups that are allowed, so that popups produced by
        # ForeignKeyRawIdWidget, on the basis of ForeignKey.limit_choices_to,
        # are allowed to work.
        for fk_lookup in model._meta.related_fkey_lookups:
            # As ``limit_choices_to`` can be a callable, invoke it here.
            if callable(fk_lookup):
                fk_lookup = fk_lookup()
            if (lookup, value) in widgets.url_params_from_lookup_dict(fk_lookup).items():
                return True

        relation_parts = []
        prev_field = None
        part = ""
        parts = lookup.split(LOOKUP_SEP)
        for part in parts:
            try:
                field = model._meta.get_field(part)
            except FieldDoesNotExist:
                # Lookups on nonexistent fields are ok, since they're ignored
                # later.
                break
            if not prev_field or (
                prev_field.is_relation
                and field not in model._meta.parents.values()
                and field is not model._meta.auto_field
                and (model._meta.auto_field is None or part not in getattr(prev_field, "to_fields", []))
                and (field.is_relation or not field.primary_key)
            ):
                relation_parts.append(part)
            if not getattr(field, "path_infos", None):
                # This is not a relational field, so further parts
                # must be transforms.
                break
            prev_field = field
            model = field.path_infos[-1].to_opts.model

        if len(relation_parts) <= 1:
            # Either a local field filter, or no fields at all.
            return True
        valid_lookups = {self.get_date_hierarchy(request, **kwargs)}
        # RemovedInDjango60Warning: when the deprecation ends, replace with:
        # for filter_item in self.get_list_filter(request):
        list_filter = self.get_list_filter(request, **kwargs) if request is not None else self.list_filter
        for filter_item in list_filter:
            if isinstance(filter_item, type) and issubclass(filter_item, SimpleListFilter):
                valid_lookups.add(filter_item.parameter_name)
            elif isinstance(filter_item, (list, tuple)):
                valid_lookups.add(filter_item[0])
            else:
                valid_lookups.add(filter_item)

        # Is it a valid relational lookup?
        return not {
            LOOKUP_SEP.join(relation_parts),
            LOOKUP_SEP.join(relation_parts + [part]),
        }.isdisjoint(valid_lookups)

    # noinspection PyUnusedLocal
    def get_list_filter(self, request, **kwargs):
        """Return a sequence containing the fields to be displayed as filters."""
        return self.list_filter

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_paginator(self, request: HttpRequest, queryset, per_page, orphans=0, allow_empty_first_page=True):
        """Return the paginator to use for this component."""
        return Paginator(queryset, per_page, orphans, allow_empty_first_page)

    # noinspection PyUnusedLocal
    def get_ordering(self, request: HttpRequest, **kwargs):
        """Return the ordering to use for this component."""
        return self.ordering or ()

    # noinspection PyUnusedLocal
    def get_search_fields(self, request: HttpRequest, **kwargs):
        """Return a sequence containing the fields to be searched."""
        return self.search_fields

    def get_search_results(
        self, request: HttpRequest, queryset: models.QuerySet, search_term: str
    ) -> Tuple[models.QuerySet, bool]:
        """Return the search result.

        Return a tuple containing a queryset to implement the search
        and a boolean indicating if the results may contain duplicates.
        """

        # Apply keyword searches.
        def construct_search(field_name):
            if field_name.startswith("^"):
                removeprefix = field_name.removeprefix("^")
                return f"{removeprefix}__istartswith"
            elif field_name.startswith("="):
                removeprefix = field_name.removeprefix("=")
                return f"{removeprefix}__iexact"
            elif field_name.startswith("@"):
                removeprefix = field_name.removeprefix("@")
                return f"{removeprefix}__search"
            # Use field_name if it includes a lookup.
            opts = self.opts
            lookup_fields = field_name.split(LOOKUP_SEP)
            # Go through the fields, following all relations.
            prev_field = None
            for path_part in lookup_fields:
                if path_part == "pk":
                    path_part = opts.pk.name
                try:
                    field = opts.get_field(path_part)
                except FieldDoesNotExist:
                    # Use valid query lookups.
                    if prev_field and prev_field.get_lookup(path_part):
                        return field_name
                else:
                    prev_field = field
                    if hasattr(field, "path_infos"):
                        # Update opts to follow the relation.
                        opts = field.path_infos[-1].to_opts
            # Otherwise, use the field with icontains.
            return f"{field_name}__icontains"

        may_have_duplicates = False
        search_fields = self.get_search_fields(request)
        if search_fields and search_term:
            orm_lookups = [construct_search(str(search_field)) for search_field in search_fields]
            term_queries = []
            for bit in smart_split(search_term):
                if bit.startswith(('"', "'")) and bit[0] == bit[-1]:
                    bit = unescape_string_literal(bit)
                or_queries = models.Q.create(
                    [(orm_lookup, bit) for orm_lookup in orm_lookups],
                    connector=models.Q.OR,
                )
                term_queries.append(or_queries)
            queryset = queryset.filter(models.Q.create(term_queries))
            may_have_duplicates |= any(lookup_spawns_duplicates(self.opts, search_spec) for search_spec in orm_lookups)
        return queryset, may_have_duplicates

    def update_render_context(self, context, **kwargs):
        """Update the context before rendering the component."""
        request = context["request"]
        cl = self.get_change_list(request, **kwargs)
        context.update({"cl": cl, "opts": self.opts})
        context.update(self.get_pagination_context(request, cl, **kwargs))
        context.update(self.get_search_context(request, cl, **kwargs))
        context.update(self.get_hierarchy_context(request, cl, **kwargs))

    # noinspection PyUnusedLocal
    def get_pagination_context(self, request: HttpRequest, cl: ChangeList, **kwargs):
        """Return the context for the pagination."""
        headers = list(result_headers(cl))
        num_sorted_fields = 0
        for h in headers:
            if h["sortable"] and h["sorted"]:
                num_sorted_fields += 1
        pagination_required = (not cl.show_all or not cl.can_show_all) and cl.multi_page
        page_range = cl.paginator.get_elided_page_range(cl.page_num) if pagination_required else []
        need_show_all_link = cl.can_show_all and not cl.show_all and cl.multi_page
        return {
            "pagination_required": pagination_required,
            "show_all_url": need_show_all_link and cl.get_query_string({self.all_var: ""}),
            "page_range": list(page_range),
            "ALL_VAR": self.all_var,
            "1": 1,
            "results": list(results(cl)),
            "result_headers": headers,
            "num_sorted_fields": num_sorted_fields,
        }

    # noinspection PyUnusedLocal
    def get_search_context(self, request: HttpRequest, cl: ChangeList, **kwargs):
        """Return the context for the search bar."""
        return {
            "show_result_count": cl.result_count != cl.full_result_count,
            "search_var": self.search_var,
            "is_popup_var": IS_POPUP_VAR,
            "is_facets_var": IS_FACETS_VAR,
        }

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_hierarchy_context(self, request: HttpRequest, cl: ChangeList, **kwargs):
        """Return the context for the hierarchy of results.

        Currently only works for date fields.
        """
        # noinspection PyTestUnpassedFixture
        field_name = cl.date_hierarchy
        if not field_name:
            return {"show_hierarchy": False}
        field = get_fields_from_path(cl.model, field_name)[-1]
        if isinstance(field, models.CharField) or isinstance(field, models.TextField):
            hierarchy_data = first_letter_hierarchy(cl)
        else:
            hierarchy_data = raw_date_hierarchy(cl)
        if hierarchy_data is not None:
            return {
                "show_hierarchy": True,
                "hierarchy_back": hierarchy_data["back"],
                "hierarchy_choices": hierarchy_data["choices"],
                "hierarchy_title": field.verbose_name,
            }
        return {"show_hierarchy": False}


class ModelAdminWrapper:
    """Wrapper around a ModelListComponent to mimic a ModelAdmin.

    This wrapper allows to add kwargs to the get_queryset method.
    ChangeList expects a ModelAdmin instance but calls the get_queryset method with
    only request arg, so we need to wrap the ModelListComponent.
    """

    def __init__(self, model_admin: ModelListComponent, kwargs):
        """Create a new wrapper."""
        self.model_admin: ModelListComponent = model_admin
        self.kwargs = kwargs

    def get_queryset(self, request):
        """Call the get_queryset method with kwargs."""
        return self.model_admin.get_queryset(request, **self.kwargs)

    def get_ordering(self, request):
        """Call the get_ordering method with kwargs."""
        return self.model_admin.get_ordering(request, **self.kwargs)

    def get_list_filter(self, request):
        """Call the get_list_filter method with kwargs."""
        return self.model_admin.get_list_filter(request, **self.kwargs)

    def get_date_hierarchy(self, request):
        """Call the get_date_hierarchy method with kwargs."""
        return self.model_admin.get_date_hierarchy(request, **self.kwargs)

    def lookup_allowed(self, lookup, value, request: HttpRequest):
        """Call the lookup_allowed method with kwargs."""
        return self.model_admin.lookup_allowed(lookup, value, request=request, **self.kwargs)

    def __getattr__(self, item):
        """Delegate all other calls to the ModelListComponent."""
        return getattr(self.model_admin, item)


def first_letter_hierarchy(cl: ChangeList):
    """Fetch all initials of a CharField."""
    # noinspection PyTestUnpassedFixture
    qs = (
        cl.queryset.annotate(hierarchy_initial=Upper(Substr(cl.date_hierarchy, 1, 1)))
        .values_list("hierarchy_initial", flat=True)
        .distinct()
    )
    values = set(qs)
    if not values:
        return None
    # noinspection PyTestUnpassedFixture
    initial_field = f"{cl.date_hierarchy}__istartswith"
    initial_value = cl.params.get(initial_field)
    if initial_value:
        cl.get_query_string(remove=[initial_field])
        back = {"link": cl.get_query_string(remove=[initial_field]), "title": _("All")}
    else:
        back = None
    return {
        "show": True,
        "back": back,
        "choices": [
            {
                "title": initial,
                "link": cl.get_query_string({initial_field: initial}) if initial != initial_value else None,
            }
            for initial in sorted(values)
        ],
    }



"""Generate the favicons."""

import importlib
import os
import pathlib
from argparse import ArgumentParser

from django.conf import settings
from django.core.management import BaseCommand
from PIL import Image


class Command(BaseCommand):
    """Generate the favicons."""

    specs = [
        ("android-chrome-192x192.png", 192, 192),
        ("android-chrome-512x512.png", 512, 512),
        ("apple-touch-icon.png", 256, 256),
        ("favicon.ico", 64, 64),
        ("favicon-16x16.png", 16, 32),
        ("favicon-32x32.png", 32, 32),
        ("mstile-150x150.png", 150, 150),
    ]

    def add_arguments(self, parser: ArgumentParser):
        """Add the arguments."""
        mod = importlib.import_module(settings.DF_MODULE_NAME)
        default = pathlib.Path(mod.__path__[0]) / "static/favicon"
        parser.add_argument("input", help="Path of the image to add.", type=pathlib.Path)
        parser.add_argument("--output", help="Destination directory.", type=pathlib.Path, default=default)

    def handle(self, *args, **options):
        """Generate the favicons from an original image to the destination directory."""
        src_path: pathlib.Path = options["input"]
        dst_dir: pathlib.Path = options["output"]
        img_src = Image.open(src_path)
        for dst_name, width, height in self.specs:
            img_dst = img_src.resize((width, height))
            dst_path: pathlib.Path = dst_dir / dst_name
            basename, sep, ext = dst_name.rpartition(".")
            os.makedirs(dst_path.parent, exist_ok=True)
            with open(dst_path, "wb") as fd:
                img_dst.save(fd, format=ext.upper())
            self.stdout.write(f"Generated {dst_path}")

        img_src.close()



"""URLs for customized views (mainly customized forms)."""

from django.conf import settings
from django.urls import include, re_path
from postman import api_urls
from postman.views import (
    ArchivesView,
    ArchiveView,
    ConversationView,
    DeleteView,
    InboxView,
    IndexView,
    MarkReadView,
    MarkUnreadView,
    MessageView,
    ReplyView,
    SentView,
    TrashView,
    UndeleteView,
    WriteView,
)

from df_site.postman.forms import HTMLAnonymousWriteForm, HTMLFullReplyForm, HTMLQuickReplyForm, HTMLWriteForm
from df_site.postman.views import MessageUpdateView

if settings.POSTMAN_I18N_URLS:
    from django.utils.translation import pgettext_lazy
else:

    def pgettext_lazy(context: str, message: str):
        """Does nothing when POSTMAN_I18N_URLS is off."""
        return message


app_name = "postman"
urlpatterns = [
    # Translators: keep consistency of the <option> parameter with the translation for 'm'
    re_path(pgettext_lazy("postman_url", r"^inbox/(?:(?P<option>m)/)?$"), InboxView.as_view(), name="inbox"),
    # Translators: keep consistency of the <option> parameter with the translation for 'm'
    re_path(pgettext_lazy("postman_url", r"^sent/(?:(?P<option>m)/)?$"), SentView.as_view(), name="sent"),
    # Translators: keep consistency of the <option> parameter with the translation for 'm'
    re_path(pgettext_lazy("postman_url", r"^archives/(?:(?P<option>m)/)?$"), ArchivesView.as_view(), name="archives"),
    # Translators: keep consistency of the <option> parameter with the translation for 'm'
    re_path(pgettext_lazy("postman_url", r"^trash/(?:(?P<option>m)/)?$"), TrashView.as_view(), name="trash"),
    re_path(
        pgettext_lazy("postman_url", r"^write/(?:(?P<recipients>[^/#]+)/)?$"),
        WriteView.as_view(form_classes=(HTMLWriteForm, HTMLAnonymousWriteForm)),
        name="write",
    ),
    re_path(
        pgettext_lazy("postman_url", r"^reply/(?P<message_id>[\d]+)/$"),
        ReplyView.as_view(form_class=HTMLFullReplyForm),
        name="reply",
    ),
    re_path(
        pgettext_lazy("postman_url", r"^view/(?P<message_id>[\d]+)/$"),
        MessageView.as_view(form_class=HTMLQuickReplyForm),
        name="view",
    ),
    # Translators: 't' stands for 'thread'
    re_path(
        pgettext_lazy("postman_url", r"^view/t/(?P<thread_id>[\d]+)/$"),
        ConversationView.as_view(form_class=HTMLQuickReplyForm),
        name="view_conversation",
    ),
    re_path(pgettext_lazy("postman_url", r"^archive/$"), ArchiveView.as_view(), name="archive"),
    re_path(pgettext_lazy("postman_url", r"^delete/$"), DeleteView.as_view(), name="delete"),
    re_path(pgettext_lazy("postman_url", r"^undelete/$"), UndeleteView.as_view(), name="undelete"),
    re_path(pgettext_lazy("postman_url", r"^mark-read/$"), MarkReadView.as_view(), name="mark-read"),
    re_path(pgettext_lazy("postman_url", r"^mark-unread/$"), MarkUnreadView.as_view(), name="mark-unread"),
    # this view is not a part of the original postman views
    re_path(pgettext_lazy("postman_url", r"^mark-message/$"), MessageUpdateView.as_view(), name="update-message"),
    re_path(r"^$", IndexView.as_view()),
    re_path(pgettext_lazy("postman_url", r"^api/"), include(api_urls, namespace="api")),
]



"""Check if a response from a Django view meets the expectations.

These expectations can be as low as a simple HTTP code.
"""

import json
import traceback
from html.parser import HTMLParser
from typing import Callable, Dict, List, Optional, Set, Tuple, Type, Union
from unittest import TestCase

from django.contrib.auth.models import AbstractUser
from django.core.exceptions import PermissionDenied
from django.core.files.uploadedfile import SimpleUploadedFile
from django.http import Http404, HttpRequest, HttpResponse, JsonResponse
from django.template.response import TemplateResponse
from django.urls import reverse


class FormHTMLParser(HTMLParser):
    """Fetch all field names in a HTML document."""

    def __init__(
        self,
        form_id: Optional[str] = None,
        convert_charrefs=True,
        ignored_fields: Optional[Set[str]] = None,
        ignored_types: Optional[Set[str]] = None,
    ):
        """Initialize the parser."""
        super().__init__(convert_charrefs=convert_charrefs)
        self.form_id = form_id
        self.form_fields: Set[str] = set()
        self.is_in_selected_form = False
        if ignored_fields is None:
            ignored_fields = {"csrfmiddlewaretoken"}
        if ignored_types is None:
            ignored_types = {"submit", "reset", "button", "image"}
        self.ignored_fields = ignored_fields
        self.ignored_types = ignored_types

    def reset(self, form_id=None):
        """Reset the parser."""
        super().reset()
        self.form_fields: Set[str] = set()
        self.is_in_selected_form = False

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, str]]):
        """Check if we are in the form and if we look at a field."""
        attrs = dict(attrs)
        if tag == "form" and (self.form_id is None or attrs.get("id") == self.form_id):
            self.is_in_selected_form = True
        if self.is_in_selected_form and tag in {"input", "select", "textarea"} and "name" in attrs:
            if attrs["name"] in self.ignored_fields:
                return
            elif tag == "input" and attrs.get("type") in self.ignored_types:
                return
            self.form_fields.add(attrs["name"])

    def handle_endtag(self, tag):
        """Check if we are still in the form."""
        if tag == "form":
            self.is_in_selected_form = False

    def error(self, message):
        """Does nothing when there is an error."""
        pass


class RequestTester:
    """Check if the response of a Django view meets the expectation."""

    def __init__(
        self,
        expected_response: Union[int, Exception, Type[Exception], HttpResponse, callable] = 200,
        method="get",
        form_data: Union[Dict[str, str], Callable] = None,
        get_data: Union[Dict[str, str], Callable] = None,
        post_data: Union[bytes, Dict[str, str], Callable] = None,
        files_data: Dict[str, SimpleUploadedFile] = None,
        raise_error: bool = False,
        reverse_kwargs: Union[Dict, Callable] = None,
        request_kwargs: Dict = None,
        view_kwargs: Union[Dict, Callable] = None,
        validators=None,
        headers: Optional[Dict[str, str]] = None,
        **response_attr,
    ):
        """Initialize the object."""
        self.method = method.lower()
        self.form_data = form_data
        self.headers = headers or {}
        self.post_data = post_data
        self.files_data = files_data or {}
        self.expected_response = expected_response
        self.raise_error = raise_error
        self.request_kwargs = request_kwargs or {}  # extra kwargs for building the HttpRequest
        self.reverse_kwargs = (
            reverse_kwargs or {}
        )  # extra kwargs provided to the callable and to the `reverse` function
        self.view_kwargs = view_kwargs or {}  # extra kwargs provided to the callabble
        self.get_data = get_data or {}  # provide the GET QueryDict of the created HttpRequest
        self.validators = validators or []  # list of callable(testCase, HttpResponse)
        self.response_attr = response_attr  # extra response attrs
        self.request = None

    @staticmethod
    def callable_value(value, test_case: TestCase, complete_url_name: str, user: AbstractUser):
        """Call the callable if possible, otherwise return the value."""
        if callable(value):
            return value(test_case, complete_url_name, user)
        return value

    # noinspection PyUnusedLocal
    def get_request_kwargs(self, test_case: TestCase, complete_url_name: str, user: AbstractUser) -> Dict:
        """Build all kwargs provided to the request factory."""
        merged_request_kwargs = {}
        merged_request_kwargs.update(self.request_kwargs)
        if self.form_data:
            merged_request_kwargs["form_data"] = self.callable_value(self.form_data, test_case, complete_url_name, user)
        if self.get_data:
            merged_request_kwargs["get_data"] = self.callable_value(self.get_data, test_case, complete_url_name, user)
        if self.post_data:
            merged_request_kwargs["post_data"] = self.callable_value(self.post_data, test_case, complete_url_name, user)
        if self.headers:
            merged_request_kwargs["headers"] = self.headers
        return merged_request_kwargs

    def get_request_method(self, test_case: TestCase, complete_url_name: str, user: AbstractUser) -> str:
        """Return the HTTP method."""
        return self.callable_value(self.method, test_case, complete_url_name, user)

    def get_view_kwargs(self, test_case: TestCase, complete_url_name: str, user: AbstractUser) -> Dict:
        """Return extra kwargs provided to the view."""
        return self.callable_value(self.view_kwargs, test_case, complete_url_name, user)

    def get_reverse_kwargs(self, test_case: TestCase, complete_url_name: str, user: AbstractUser) -> Dict:
        """Return the kwargs used for reversing the URL."""
        return self.callable_value(self.reverse_kwargs, test_case, complete_url_name, user)

    def get_http_request(self, test_case, view_name: str, reverse_kwargs: Dict, user: AbstractUser) -> HttpRequest:
        """Return the HttpRequest."""
        if view_name:
            url = reverse(view_name, kwargs=reverse_kwargs)
        else:
            url = "/unnamed/view"
        method = self.get_request_method(test_case, view_name, user)
        request_kwargs = self.get_request_kwargs(test_case, view_name, user)
        request = test_case.create_request(url, method=method, **request_kwargs)
        if self.files_data:
            request.FILES.update(self.files_data)
        request.user = user
        return request

    def evaluate(
        self,
        test_case,
        view_name: str,
        view: callable,
        reverse_kwargs: Dict,
        view_kwargs: Dict,
        user: AbstractUser,
    ) -> Optional[Exception]:
        """Evaluate a single view with the provided reverse kwargs and view kwargs.

        :param test_case: the test case evaluating this function (maybe some attributes are required)
        :param view_name: the name of the tested view
        :param view: the tested view itself
        :param reverse_kwargs: kwargs used for the `reverse` function (updated by `get_reverse_kwargs`)
        :param view_kwargs: extra args passed to the view (updated by `get_view_kwargs`)
        :param user: the tested permission set
        """
        merged_reverse_kwargs = {}
        merged_reverse_kwargs.update(reverse_kwargs)
        merged_reverse_kwargs.update(self.get_reverse_kwargs(test_case, view_name, user))

        request = self.get_http_request(test_case, view_name, merged_reverse_kwargs, user)
        self.request = request

        merged_view_kwargs = merged_reverse_kwargs
        merged_view_kwargs.update(view_kwargs)
        merged_view_kwargs.update(self.get_view_kwargs(test_case, view_name, user))
        if self.raise_error:
            self.check_response(test_case, request, view, merged_view_kwargs, user)
        else:
            try:
                self.check_response(test_case, request, view, merged_view_kwargs, user)
            except Exception as e:
                if not isinstance(e, PermissionDenied) and not isinstance(e, Http404):
                    traceback.print_exc()
                return e

    def check_response(
        self,
        test_case: TestCase,
        request: HttpRequest,
        view: callable,
        view_kwargs: Dict,
        user: AbstractUser,
    ):
        """Check if the response meets the expectation."""
        expected_response = self.expected_response
        # noinspection PyUnusedLocal
        user = user
        if isinstance(expected_response, type) and issubclass(expected_response, Exception):
            test_case.assertRaises(expected_response, lambda: view(request, **view_kwargs))
            return
        http_response = view(request, **view_kwargs)
        if isinstance(http_response, TemplateResponse):
            http_response.render()
        for validator in self.validators:
            validator(test_case, http_response)
        if isinstance(expected_response, int):
            if expected_response != http_response.status_code:
                # do not use test_case.assertEqual to have a more explicit error
                msg = f"HTTP {http_response.status_code} (while expecting {expected_response})"
                msg += self.get_description(request)
                raise AssertionError(msg)
        elif isinstance(expected_response, HttpResponse):
            if expected_response.status_code != http_response.status_code:
                # do not use test_case.assertEqual to have a more explicit error
                msg = f"HTTP {http_response.status_code} (while expecting {expected_response.status_code})"
                msg += self.get_description(request)
                raise AssertionError(msg)
            test_case.assertIsInstance(http_response, expected_response.__class__)
        elif callable(expected_response):
            expected_response(test_case, http_response)
        else:
            raise ValueError(f"unknown type of expected response: {expected_response!r}")
        for attr_name, attr_value in self.response_attr.items():
            if isinstance(attr_value, type):
                test_case.assertIsInstance(attr_value, getattr(http_response, attr_name))
            else:
                test_case.assertEqual(attr_value, getattr(http_response, attr_name))

    def get_description(self, request: HttpRequest) -> str:
        """Return a string representation of a RequestTester."""
        msg = f" url = {request.path}"
        if self.method != "get":
            msg += f" method={self.method}"
        if self.headers:
            msg += f" headers={self.headers}"
        if self.post_data:
            post_data = str(self.post_data)[:30]
            msg += f" post_data={post_data}"
        if self.form_data or self.get_data:
            post_data = str(self.form_data)[:30]
            msg += f" form_data={post_data}"
        if self.get_data:
            post_data = str(self.get_data)[:30]
            msg += f" get_data={post_data}"
        return msg


class HTMLFormRequestTester(RequestTester):
    """Validate a form in the HTML response, for example to check if required fields are present."""

    def __init__(
        self,
        required_fields: Set[str],
        *args,
        form_id: Optional[str] = None,
        ignored_fields: Optional[Set[str]] = None,
        ignored_types: Optional[Set[str]] = None,
        **kwargs,
    ):
        """Initialize the object with expected values."""
        self.required_fields = required_fields
        kwargs.setdefault("validators", [])
        kwargs["validators"].append(self.check_required_fields)
        self.form_id = form_id
        self.ignored_fields = ignored_fields
        self.ignored_types = ignored_types
        super().__init__(*args, **kwargs)

    def check_required_fields(self, test_case: TestCase, response: HttpResponse):
        """Check if the response contains the expected form values."""
        parser = FormHTMLParser(
            form_id=self.form_id,
            ignored_types=self.ignored_types,
            ignored_fields=self.ignored_fields,
        )
        parser.feed(response.content.decode())
        test_case.assertEqual(self.required_fields, parser.form_fields)


class JsonValidator:
    """Validate a JSONResponse."""

    def __init__(self, value):
        """Initialize the object with the expected JSON."""
        self.value = value

    def __call__(self, test_case: TestCase, response: JsonResponse):
        """Check the content of a JsonResponse."""
        content = json.loads(response.content)
        if self.value != content:
            print(content)
        test_case.assertEqual(self.value, content)



"""Component and generic View for displaying the details of a model instance."""

from typing import Dict, List, Optional, Type, Union

from df_websockets.tasks import set_websocket_topics
from django import forms
from django.contrib.admin import helpers
from django.contrib.admin.utils import flatten_fieldsets
from django.core.exceptions import ValidationError
from django.db import models
from django.forms import fields_for_model
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import DetailView

from df_site.components.base import ModelComponent


class ModelDetailComponent(ModelComponent):
    """A component that displays the details of a model instance."""

    model: Type[models.Model]

    def __init__(
        self,
        model: Type[models.Model],
        base_template: str = "detail.html",
        inlines: List = None,
        fields=None,
        exclude=None,
        fieldsets=None,
        filter_vertical=(),
        filter_horizontal=(),
    ):
        """Initialize the component."""
        super().__init__(model, base_template)
        self.inlines = inlines or []
        self.fields = fields
        self.exclude = exclude
        self.fieldsets = fieldsets
        self.filter_vertical = filter_vertical
        self.filter_horizontal = filter_horizontal
        self.opts = model._meta

    def get_queryset(self, request):
        """Return a QuerySet of all model instances that can be viewed."""
        # noinspection PyProtectedMember
        return self.model._default_manager.get_queryset()

    def get_fields(self, request, obj=None):
        """Return the list of selected fields."""
        if self.fields:
            return self.fields
        # _get_form_for_get_fields() is implemented in subclasses.
        fields = fields_for_model(self.model)
        return [*fields]

    def get_fieldsets(self, request, obj=None):
        """Return the specified fieldsets."""
        if self.fieldsets:
            return self.fieldsets
        return [(None, {"fields": self.get_fields(request, obj)})]

    def get_inlines(self, request, obj):
        """Hook for specifying custom inlines."""
        return self.inlines

    def get_object(self, request, object_id, from_field=None):
        """Return an instance matching the field and value provided.

        The primary key is used if no field is provided. Return ``None`` if no match is
        found or the object_id fails validation.
        """
        queryset = self.get_queryset(request)
        model = queryset.model
        field = self.opts.pk if from_field is None else self.opts.get_field(from_field)
        try:
            object_id = field.to_python(object_id)
            return queryset.get(**{field.name: object_id})
        except (model.DoesNotExist, ValidationError, ValueError):
            return None

    def update_render_context(self, context, **kwargs):
        """Update the context before rendering the component.

        The displayed object can be specified either in the context as "object",
         in the kwargs as "obj" or its ID as "object_id" in kwargs.
        """
        request = context["request"]
        if "object" in context:
            obj = context["object"]
        elif "obj" in kwargs:
            obj = kwargs["obj"]
        else:
            object_id = kwargs.get("object_id")
            from_field = kwargs.get("from_field")
            obj = self.get_object(request, object_id=object_id, from_field=from_field)
        fieldsets = self.get_fieldsets(request, obj)
        readonly_fields = flatten_fieldsets(fieldsets)

        class AdminForm(forms.ModelForm):
            class Meta:
                model = self.model
                fields = []

        form = AdminForm(instance=obj)
        admin_form = helpers.AdminForm(
            form,
            list(fieldsets),
            {},
            readonly_fields=readonly_fields,
            model_admin=self,
        )
        context["adminform"] = admin_form


class ModelDetailView(DetailView):
    """A view that displays the details of a model instance."""

    component: ModelDetailComponent
    model: Type[models.Model]
    component_template: str = "detail.html"
    inlines: List = None
    fields = None
    exclude = None
    fieldsets = None
    filter_vertical = ()
    filter_horizontal = ()

    @classmethod
    def as_view(cls, **initkwargs):
        """Main entry point for a request-response process."""
        cls.component = ModelDetailComponent(
            model=cls.model,
            base_template=cls.component_template,
            inlines=cls.inlines,
            fields=cls.fields,
            exclude=cls.exclude,
            fieldsets=cls.fieldsets,
            filter_vertical=cls.filter_vertical,
            filter_horizontal=cls.filter_horizontal,
        )
        return super().as_view(**initkwargs)

    def get_template_names(self):
        """Return a list of template names to be used for the request."""
        if self.template_name:
            return [self.template_name]
        return [
            f"df_site/{self.component.opts.app_label}/{self.component.opts.model_name}/detail.html",
            f"df_site/{self.component.opts.app_label}/detail.html",
            "df_site/detail.html",
        ]

    def get_breadcrumb(self) -> List[Dict[str, Union[str, bool]]]:
        """Return the breadcrumb for the view.

        The breadcrumb is a list of dictionaries with the following keys:
        * `link`: URL to link to
        * `title`: Text to display
        * `active`: boolean, element is active, meaning that the link is not displayed
        """
        return [
            {"link": reverse("index"), "title": _("Home"), "active": False},
            {"link": None, "title": str(self.object), "active": True},
        ]

    def get_context_data(self, **kwargs):
        """Get the context data for the view."""
        context = super().get_context_data(**kwargs)
        context["detail_component"] = self.component
        context["detail_breadcrumb"] = self.get_breadcrumb()
        context["PAGE_TITLE"] = self.get_page_title()
        context["PAGE_DESCRIPTION"] = self.get_page_description()
        context["PAGE_URL"] = self.get_page_url(context)
        set_websocket_topics(self.request, self.object)
        return context

    def get_page_url(self, context) -> Optional[str]:
        """Return the URL of the current page."""
        if hasattr(self.object, "get_absolute_url"):
            return self.object.get_absolute_url()

    def get_page_title(self) -> Optional[str]:
        """Return the title of the current page."""
        return str(self.object)

    def get_page_description(self) -> Optional[str]:
        """Return the description of the current page."""
        # noinspection PyProtectedMember
        return self.object._meta.verbose_name



"""List of URLs for the df_site app."""

from django.conf import settings
from django.urls import include, path
from django.utils.module_loading import import_string
from django.views.generic import RedirectView
from django_prometheus import exports

from df_site.views import (
    BrowserConfigView,
    HumansTxtView,
    SecurityTxtView,
    csp_report_view,
    security_gpg_view,
    site_webmanifest_view,
    thumbnail_view,
)

urlpatterns = [
    path(
        ".well-known/change-password",
        RedirectView.as_view(pattern_name="account_change_password", permanent=True),
        name="well-known-change-password",
    ),
    path(".well-known/security.txt", SecurityTxtView.as_view(), name="well-known-security"),
    path(".well-known/humans.txt", HumansTxtView.as_view(), name="well-known-humans"),
    path(".well-known/gpg.txt", security_gpg_view, name="well-known-gpg"),
    path("site.webmanifest", site_webmanifest_view, name="site_webmanifest"),
    path("browserconfig.xml", BrowserConfigView.as_view(), name="browserconfig"),
    path("metrics", exports.ExportToDjangoView, name="prometheus-django-metrics"),
    path(settings.CSP_REPORT_URI[1:], csp_report_view, name="csp_report"),
    path("users/", include("df_site.users.urls", namespace="users")),
    path("thumbnails/<path:path>", thumbnail_view, name="thumbnails"),
    path("messages/", include("df_site.postman.urls", namespace="postman")),
    path("cookies/", include("cookie_consent.urls")),
    path(
        "upload_file/",
        import_string(settings.CK_EDITOR_5_UPLOAD_FILE_VIEW),
        name=settings.CK_EDITOR_5_UPLOAD_FILE_VIEW_NAME,
    ),
]



"""Define a HTML component class, a elaborated piece of template.

A component should be instantiated once and not have any request-specific attribute.
Request-specific data should be passed as arguments to methods.
"""

from typing import List, Optional, Type

from django.db import models
from django.http import HttpRequest


class Component:
    """Base component class."""

    template_name: str = "components/base.html"

    def render(self, context, **kwargs) -> str:
        """Render the component in a HTML template.

        Context the complete template context.
        Kwargs is a dictionary of data provided to the template tag.
        """
        self.update_render_context(context, **kwargs)
        names = self.get_template_names()
        template = context.template.engine.select_template(names)
        return template.render(context)

    def get_template_names(self) -> List[str]:
        """Return a list of template names to be used for the request."""
        return [self.template_name]

    def update_render_context(self, context, **kwargs):
        """Update the context before rendering the component.

        Context the complete template context.
        Kwargs is a dictionary of data provided to the template tag.
        """
        pass


class ModelComponent(Component):
    """Base component class for model-related components."""

    template_name: Optional[str] = None

    def __init__(self, model: Type[models.Model], base_template: str):
        """Initialize the component."""
        super().__init__()
        self.model: Type[models.Model] = model
        # noinspection PyProtectedMember
        self.opts = model._meta
        self.base_template: str = base_template

    @property
    def id_prefix(self) -> str:
        """Propose a prefix for the HTML IDs linked to this component."""
        return f"{self.opts.app_label}_{self.opts.model_name}"

    # noinspection PyUnusedLocal
    def get_base_queryset(self, request: HttpRequest, **kwargs) -> models.QuerySet:
        """Return the base queryset to use for this component.

        Any extra filter based on kwargs should be applied here.
        """
        # noinspection PyProtectedMember
        return self.model._default_manager.get_queryset()

    # noinspection PyMethodMayBeStatic
    def get_empty_value_display(self) -> str:
        """Return the value to display for an empty field."""
        return "-"

    def get_template_names(self) -> List[str]:
        """Return a list of template names to be used for the request."""
        if self.template_name:
            return [self.template_name]
        return [
            f"df_components/{self.opts.app_label}/{self.opts.model_name}/{self.base_template}",
            f"df_components/{self.opts.app_label}/{self.base_template}",
            f"df_components/{self.base_template}",
        ]



"""Admin classes for the df_site app."""

from django.conf import settings
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.utils.translation import gettext as _

from df_site.models import AlertRibbon, PreferencesUser


@admin.register(AlertRibbon)
class AlertRibbonAdmin(admin.ModelAdmin):
    """Admin class for the alert ribbon model."""

    list_display = ("message", "color", "start_date", "end_date", "is_active")
    list_filter = ("color", "start_date", "end_date", "is_active", "position")
    search_fields = ("summary",)
    fields = ["summary", "url", "message", "color", "start_date", "end_date", "is_active", "position"]


if "allauth.account" in settings.INSTALLED_APPS:
    from allauth.account.models import EmailAddress

    class EmailAddressInline(admin.TabularInline):
        """Inline for the email address model."""

        model = EmailAddress
        extra = 0
else:
    EmailAddressInline = None

if "allauth.mfa" in settings.INSTALLED_APPS:
    from allauth.mfa.models import Authenticator

    class AuthenticatorInline(admin.TabularInline):
        """Inline for the authenticator model."""

        model = Authenticator
        extra = 0
        fields = ["type", "created_at", "last_used_at"]
        readonly_fields = ["type", "created_at", "last_used_at"]

        def has_add_permission(self, request, obj):
            """Return False to prevent adding new authenticators."""
            return False
else:
    AuthenticatorInline = None

if "allauth.socialaccount" in settings.INSTALLED_APPS:
    from allauth.socialaccount.models import SocialAccount

    class SocialAccountInline(admin.TabularInline):
        """Inline for the social account model."""

        model = SocialAccount
        extra = 0
        fields = ["provider", "last_login", "date_joined"]
        readonly_fields = ["provider", "last_login", "date_joined"]

        def has_add_permission(self, request_, obj_):
            """Return False to prevent adding new social accounts."""
            return False
else:
    SocialAccountInline = None

if "allauth.usersessions" in settings.INSTALLED_APPS:
    from allauth.usersessions.models import UserSession

    class UserSessionInline(admin.TabularInline):
        """Inline for the user session model."""

        model = UserSession
        extra = 0
        fields = [
            "created_at",
            "ip",
            "last_seen_at",
        ]
        readonly_fields = [
            "created_at",
            "ip",
            "last_seen_at",
        ]

        def has_add_permission(self, request_, obj):
            """Return False to prevent adding new user sessions."""
            return False
else:
    UserSessionInline = None


@admin.register(PreferencesUser)
class PreferencesUserAdmin(UserAdmin):
    """Admin class for the preferences user model."""

    inlines = []
    list_display = ("username", "email", "first_name", "last_name", "is_staff", "date_joined")
    list_filter = ("is_staff", "is_superuser", "is_active", "date_joined", "groups")
    readonly_fields = ["last_login", "date_joined"]
    fieldsets = (
        (None, {"fields": ("username", "password")}),
        (
            _("Personal info"),
            {"fields": ("first_name", "last_name", "email", "color_theme", "display_online", "email_notifications")},
        ),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
            },
        ),
        (_("Important dates"), {"fields": (("last_login", "date_joined"),)}),
    )

    def get_inlines(self, request, obj):
        """Return inlines excluding those that correspond to apps not installed."""
        inlines = [EmailAddressInline, AuthenticatorInline, SocialAccountInline, UserSessionInline]
        if obj is None:
            return []
        inlines = [x for x in inlines if x is not None]
        return inlines



"""Views for the df_site app."""

import datetime
import json
import logging

from django.conf import settings
from django.contrib import messages
from django.http import Http404, HttpRequest, HttpResponse, JsonResponse
from django.templatetags.static import static
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import TemplateView
from django.views.static import serve

from df_site.templatetags.df_site import abs_url
from df_site.templatetags.images import CachedImage

logger = logging.getLogger(__name__)


def site_webmanifest_view(request: HttpRequest) -> HttpResponse:
    """Generate a site.webmanifest view."""
    result = {
        "name": settings.DF_SITE_TITLE,
        "short_name": settings.DF_SITE_TITLE,
        "icons": [
            {"src": static("favicon/android-chrome-192x192.png"), "sizes": "192x192", "type": "image/png"},
            {"src": static("favicon/android-chrome-512x512.png"), "sizes": "512x512", "type": "image/png"},
        ],
        "theme_color": settings.DF_ANDROID_THEME_COLOR,
        "background_color": settings.DF_ANDROID_BACKGROUND_COLOR,
        "display": "standalone",
    }
    return JsonResponse(result)


@csrf_exempt
def csp_report_view(request: HttpRequest) -> HttpResponse:
    """View to receive CSP reports, displaying them to the user in DEBUG mode."""
    logger.info("CSP report: %s", request.body)
    if settings.DEBUG:
        try:
            content = json.loads(request.body)
            csp_report = content["csp-report"]
            msg = (
                f"<strong>CSP violation</strong> on this <a href='{csp_report['document-uri']}'>page</a>:"
                f" {csp_report['effective-directive']} forbids the use of"
                f" '{csp_report['blocked-uri']}' URIs."
            )
            messages.error(request, mark_safe(msg))  # noqa
        except ValueError:
            pass
    return HttpResponse(status=204)


class BrowserConfigView(TemplateView):
    """View for the browserconfig.xml file."""

    template_name = "favicon/browserconfig.xml"
    content_type = "application/xml"


def security_gpg_view(request: HttpRequest) -> HttpResponse:
    """Return the GPG key to communicate with about security problems."""
    if settings.DF_SITE_SECURITY_GPG_CONTENT:
        raise Http404
    response = HttpResponse(settings.DF_SITE_SECURITY_GPG_CONTENT, content_type="text/plain")
    response["Content-Disposition"] = 'attachment; filename="gpg.txt"'
    return response


class SecurityTxtView(TemplateView):
    """View for the ./well-known/security.txt file."""

    template_name = "df_site/security.txt"
    content_type = "plain/text"

    def get_context_data(self, **kwargs):
        """Return the context data for the view."""
        context = super().get_context_data(**kwargs)
        context["security_txt_path"] = abs_url(reverse("well-known-security"))
        context["security_email"] = settings.DF_SITE_SECURITY_EMAIL
        context["security_language_code"] = settings.DF_SITE_SECURITY_LANGUAGE_CODE
        gpg_key = None
        if settings.DF_SITE_SECURITY_GPG_CONTENT:
            gpg_key = abs_url(reverse("well-known-gpg"))
        context["security_gpg_key"] = gpg_key
        now = datetime.datetime.now(tz=datetime.UTC)
        now = now - datetime.timedelta(microseconds=now.microsecond)
        context["security_expires"] = now + datetime.timedelta(days=30)
        return context


class HumansTxtView(TemplateView):
    """View for the ./well-known/humans.txt file."""

    template_name = "df_site/humans.txt"
    content_type = "plain/text"

    def get_context_data(self, **kwargs):
        """Return the context data for the view."""
        context = super().get_context_data(**kwargs)
        now = datetime.datetime.now(tz=datetime.UTC)
        now = now - datetime.timedelta(microseconds=now.microsecond)
        context["humans_description"] = settings.DF_SITE_DESCRIPTION
        context["humans_social_networks"] = settings.DF_SITE_SOCIAL_NETWORKS
        context["humans_keywords"] = settings.DF_SITE_KEYWORDS
        context["humans_update"] = now
        return context


def thumbnail_view(request: HttpRequest, path: str) -> HttpResponse:
    """Return a thumbnail image."""
    img = CachedImage.from_target_path(path)
    if not img.src_storage_obj.exists(path):
        img.process()
    return serve(request, path=path, document_root=settings.STATIC_ROOT)


#
